/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include "alsamidi.h"
#include "track.h"
#include "tracker.h"

#include <iostream>
#include <klocale.h>
#include <sys/poll.h>
#include <qtimer.h>

using namespace std;

/**
 * @short Initialize the midi interface with ALSA
 */
Midi::Midi(SysExBag *bag, QObject *parent,const char *name): QObject(parent,name){
  int err;
  err = snd_seq_open(&handle, "default", SND_SEQ_OPEN_DUPLEX, SND_SEQ_NONBLOCK);
  if (err < 0){
    cerr<<"Error initialising MIDI"<<endl;
    exit(1);
  }
  snd_seq_set_client_name(handle, name);
  currentPort=addInputPort(i18n("Current"));
  inBlockMode=false;
  sysexbag=bag;
  
  QTimer *t=new QTimer(this,"ALSA ports timer");
  t->start(5*1000,false);
  QObject::connect(t,SIGNAL(timeout()),this,SLOT(recheckALSAPorts()));
  recheckALSAPorts();
}

/**
 * @short Deletes the interface.
 */
Midi::~Midi(){
  snd_seq_close(handle);
  // Delete lists
  ALSAPort *port;
  for (port=outputLastPorts.first();port;port=outputLastPorts.next())
    delete port;
}


/**
 * @short Adds a new port to the interface. An output port.
 */
int Midi::addPort(const char *name){
  //  QMutexLocker l(&lock);
  int port;
  port=snd_seq_create_simple_port(handle, name,
				    SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
				    SND_SEQ_PORT_TYPE_MIDI_GENERIC);
  if (port>=0)
    portList.append(port);
  return port;
}
/**
 * @short Adds a new port to the interface. An input port.
 */
int Midi::addInputPort(const char *name){
  //  QMutexLocker l(&lock);
  int ok;
  ok=snd_seq_create_simple_port(handle, name,
				SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
				SND_SEQ_PORT_TYPE_APPLICATION);
  return ok;
}
/**
 * @short Deletes a port from the map.
 */
void Midi::delPort(int id){
  //  QMutexLocker l(&lock);
  snd_seq_delete_simple_port(handle, id);
  portList.remove(id);
}


/**
 * @short Given a port number and a midinote, it plays it.
 */
void Midi::playNote(const MidiNote &note, int MIDIChannel, int port){
  QMutexLocker l(&lock);
  snd_seq_event_t ev;

  switch(note.note()){
  case MidiNote::NoNote:
  case MidiNote::StopNote:
    break;
  default:
    snd_seq_ev_clear(&ev);
    snd_seq_ev_set_source(&ev, port);
    snd_seq_ev_set_subs(&ev);
    snd_seq_ev_set_direct(&ev);
    if (note.velocity()==0){
      snd_seq_ev_set_noteoff(&ev, MIDIChannel-1, note.note(), note.velocity());
    }
    else{
      snd_seq_ev_set_noteon(&ev, MIDIChannel-1, note.note(), note.velocity());
    }
    snd_seq_event_output(handle, &ev);
    snd_seq_drain_output(handle);
    
  }

}

/**
 * @short Plays a controller event
 */
void Midi::playController(char controllerId, char val, int MIDIChannel, int port){
  QMutexLocker l(&lock);
  snd_seq_event_t ev;

  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev, port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);

  snd_seq_ev_set_controller(&ev, MIDIChannel-1, controllerId, val);

  snd_seq_event_output(handle, &ev);
  snd_seq_drain_output(handle);
}

/**
 * @short Sends a SysEx controller
 */
void Midi::sendSysEx(const char *sysex, unsigned int len, int , int port){
  QMutexLocker l(&lock);
  snd_seq_event_t ev;

  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev,port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);

  snd_seq_ev_set_sysex(&ev, len, (char*)sysex);

  snd_seq_event_output(handle, &ev);
  snd_seq_drain_output(handle);
}

/**
 * @short Sends a pitch bend message.
 */
void Midi::pitchBend(float value, int MIDIChannel, int port){
  QMutexLocker l(&lock);
  snd_seq_event_t ev;

  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev,port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);

  snd_seq_ev_set_pitchbend(&ev, MIDIChannel-1, int(value*8191));

  snd_seq_event_output(handle, &ev);
  snd_seq_drain_output(handle);
}

/**
 * @short Sends a Change instrument message.
 */
void Midi::programChange(int prog, int MIDIChannel, int port){
  QMutexLocker l(&lock);
  snd_seq_event_t ev;

  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_source(&ev,port);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);

  snd_seq_ev_set_pgmchange(&ev, MIDIChannel-1, prog);

  snd_seq_event_output(handle, &ev);
  snd_seq_drain_output(handle);
}

/**
 * @short Sends start play command
 */
void Midi::sendStartPlay(){
  QMutexLocker l(&lock);
  unsigned int i;
  snd_seq_event_t ev;
  
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);
  
  ev.type=SND_SEQ_EVENT_START;
  
  for (i=0;i<portList.count();i++){
    snd_seq_ev_set_source(&ev,portList[i]);
    snd_seq_event_output(handle, &ev);
  }
  
  snd_seq_drain_output(handle);
}

/**
 * @short Sends stop play command
 */
void Midi::sendStopPlay(){
  QMutexLocker l(&lock);
  unsigned int i;
  snd_seq_event_t ev;
  
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);
  
  ev.type=SND_SEQ_EVENT_STOP;
  
  for (i=0;i<portList.count();i++){
    snd_seq_ev_set_source(&ev,portList[i]);
    snd_seq_event_output(handle, &ev);
  }
  
  snd_seq_drain_output(handle);
}

/**
 * @short Sends MIDI clock signal
 */
void Midi::sendMIDIClock(){
  QMutexLocker l(&lock);
  unsigned int i;
  snd_seq_event_t ev;
  
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_subs(&ev);
  snd_seq_ev_set_direct(&ev);
  
  ev.type=SND_SEQ_EVENT_CLOCK;
  
  for (i=0;i<portList.count();i++){
    snd_seq_ev_set_source(&ev,portList[i]);
    snd_seq_event_output(handle, &ev);
  }
  
  bool oldBlockMode=inBlockMode;
  setBlockMode(true);
  snd_seq_drain_output(handle);
  setBlockMode(oldBlockMode);
}

/**
 * @short Changes the port name to name
 */
void Midi::setPortName(int port, const char *name){
  QMutexLocker l(&lock);
  snd_seq_port_info_t *info;
  int nok;

  info=(snd_seq_port_info_t*)malloc(snd_seq_port_info_sizeof());
  if (info==NULL) return;

  nok=snd_seq_get_port_info(handle, port, info);
  if (nok) return;

  snd_seq_port_info_set_name(info, name);

  nok=snd_seq_set_port_info(handle, port, info);
  if (nok) return;

  free(info);
}

/**
 * @short Decodes an alsa event and process it
 */
void Midi::decodeAlsaEvent(snd_seq_event_t *ev, Tracker *tracker){
  switch(ev->type){
  case SND_SEQ_EVENT_NOTEON:
    if (ev->data.note.velocity!=0)
      tracker->addNote(MidiNote(ev->data.note.note,ev->data.note.velocity));
    /*
      else
      tracker->addNote(MidiNote(MidiNote::StopNote,ev->data.note.velocity));
      break;
      case SND_SEQ_EVENT_NOTEOFF:
      tracker->addNote(MidiNote(MidiNote::StopNote,ev->data.note.velocity));
    */
      break;
  case SND_SEQ_EVENT_START:
    tracker->startPlay();
    break;
  case SND_SEQ_EVENT_STOP:
    tracker->stopPlay();
    break;
  case SND_SEQ_EVENT_PORT_UNSUBSCRIBED:
    puts("MIDI diconnected");
    tracker->revertTimingMethod();
    break;
  case SND_SEQ_EVENT_CLOCK:
    tracker->midiClock();
    break;
  case SND_SEQ_EVENT_SYSEX:
    sysexbag->appendUnnamed(((char*)ev->data.ext.ptr),ev->data.ext.len);
    break;
  }
}

/**
 * @short Gets last played note on the current input
 *
 * ALSA stores events in a FIFO, so no event is lost.
 */
void Midi::updateMIDIInput(Tracker *tracker){
  snd_seq_event_t *ev;

  lock.lock();
  setBlockMode(false);
  while (snd_seq_event_input(handle, &ev)>=0){ 
    lock.unlock();
    decodeAlsaEvent(ev,tracker);
    lock.lock();
  }
  lock.unlock();
}

/**
 * @short Gets last played note on the current input
 *
 * ALSA stores events in a FIFO, so no event is lost.
 */
void Midi::updateMIDIInputBlock(Tracker *tracker){
  snd_seq_event_t *ev;
    
  //  while (snd_seq_event_input(handle, &ev)>=0){ 

  lock.lock();
  setBlockMode(true);
  snd_seq_event_input(handle, &ev);
  lock.unlock();
  decodeAlsaEvent(ev,tracker);

    //}
}

/**
 * @short Sets blocking or nonblocking method
 */
void Midi::setBlockMode(bool block){
  if (block==inBlockMode)
    return;
  snd_seq_nonblock(handle, block ? 0 : 1);
  inBlockMode=block;
}

/**
 * @short Temporized check for new ports
 *
 * It checks if new ports have been added (or names changed). If any
 * change have happened, a ALSA*PortsList() signal is emmited.
 */
void Midi::recheckALSAPorts(){

  recheckALSAOutputPorts();
  recheckALSAInputPorts();
}

/**
 * @short Temporized check for new output ports (ports to which KMiditracker outputs)
 *
 * It checks if new ports have been added (or names changed). If any
 * change have happened, a ALSAOutputPortsList() signal is emmited.
 */
void Midi::recheckALSAOutputPorts(){
  snd_seq_client_info_t *cinfo;
  snd_seq_port_info_t *pinfo;
  int count;
  QPtrList<ALSAPort> outports;

  // First, I get the current list
  snd_seq_client_info_alloca(&cinfo);
  snd_seq_port_info_alloca(&pinfo);
  snd_seq_client_info_set_client(cinfo, -1);
  while (snd_seq_query_next_client(handle, cinfo) >= 0) {
    /* reset query info */
    snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
    snd_seq_port_info_set_port(pinfo, -1);
    count = 0;
    while (snd_seq_query_next_port(handle, pinfo) >= 0) {
      count++;
      unsigned int bits=SND_SEQ_PORT_CAP_WRITE;//|SND_SEQ_PORT_CAP_SUBS_WRITE;
      if (((snd_seq_port_info_get_capability(pinfo) & (bits)) == bits) &&
	  (snd_seq_client_info_get_client(cinfo)!=snd_seq_client_id(handle))){
	outports.append(new ALSAPort(
				  snd_seq_client_info_get_client(cinfo),
				  snd_seq_port_info_get_port(pinfo),	       
				  snd_seq_client_info_get_name(cinfo),
				  snd_seq_port_info_get_name(pinfo)));
      }
    }
  }

  int changed=0;
  // Then I check if there is something new...

  // First the simplest check, and more common
  if (outputLastPorts.count()!=outports.count())
    changed=1;
  else{
    // Check names and ports
    unsigned int i;
    for (i=0;i<outputLastPorts.count();i++){
      if ((outports.at(i)->client!=outputLastPorts.at(i)->client) ||
	  (outports.at(i)->port!=outputLastPorts.at(i)->port) ||
	  (outports.at(i)->client_name!=outputLastPorts.at(i)->client_name) ||
	  (outports.at(i)->port_name!=outputLastPorts.at(i)->port_name))
	changed=1;
    }
  }


  // If something changed, I emit the signal, and prepare the old list for deletion
  if (changed){
    printf("ALSA ports changed\n");
    emit ALSAOutputPortList(outports);
    // I swap them to delete lastport instead of port
    QPtrList<ALSAPort> t;
    t=outputLastPorts;
    outputLastPorts=outports;
    outports=t;
  }
  // Delete lists
  ALSAPort *port;
  for (port=outports.first();port;port=outports.next())
    delete port;
}

/**
 * @short Temporized check for new input ports (ports to which KMiditracker inputs)
 *
 * It checks if new ports have been added (or names changed). If any
 * change have happened, a ALSAInputPortsList() signal is emmited.
 */
void Midi::recheckALSAInputPorts(){
  snd_seq_client_info_t *cinfo;
  snd_seq_port_info_t *pinfo;
  int count;
  QPtrList<ALSAPort> inports;

  // First, I get the current list
  snd_seq_client_info_alloca(&cinfo);
  snd_seq_port_info_alloca(&pinfo);
  snd_seq_client_info_set_client(cinfo, -1);
  while (snd_seq_query_next_client(handle, cinfo) >= 0) {
    /* reset query info */
    snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
    snd_seq_port_info_set_port(pinfo, -1);
    count = 0;
    while (snd_seq_query_next_port(handle, pinfo) >= 0) {
      count++;
      unsigned int bits=SND_SEQ_PORT_CAP_READ;//|SND_SEQ_PORT_CAP_SUBS_WRITE;
      if (((snd_seq_port_info_get_capability(pinfo) & (bits)) == bits) &&
	  (snd_seq_client_info_get_client(cinfo)!=snd_seq_client_id(handle))){
	inports.append(new ALSAPort(
				  snd_seq_client_info_get_client(cinfo),
				  snd_seq_port_info_get_port(pinfo),	       
				  snd_seq_client_info_get_name(cinfo),
				  snd_seq_port_info_get_name(pinfo)));
      }
    }
  }

  int changed=0;
  // Then I check if there is something new...

  // First the simplest check, and more common
  if (inputLastPorts.count()!=inports.count())
    changed=1;
  else{
    // Check names and ports
    unsigned int i;
    for (i=0;i<inputLastPorts.count();i++){
      if ((inports.at(i)->client!=inputLastPorts.at(i)->client) ||
	  (inports.at(i)->port!=inputLastPorts.at(i)->port) ||
	  (inports.at(i)->client_name!=inputLastPorts.at(i)->client_name) ||
	  (inports.at(i)->port_name!=inputLastPorts.at(i)->port_name))
	changed=1;
    }
  }


  // If something changed, I emit the signal, and prepare the old list for deletion
  if (changed){
    printf("ALSA input ports changed\n");
    emit ALSAInputPortList(inports);
    // I swap them to delete lastport instead of port
    QPtrList<ALSAPort> t;
    t=inputLastPorts;
    inputLastPorts=inports;
    inports=t;
  }
  // Delete lists
  ALSAPort *port;
  for (port=inports.first();port;port=inports.next())
    delete port;
}

/**
 * @short Reemits the ALSAPortList() signal. 
 *
 * Usefull when a new track is added.
 */
void Midi::resendPortList(){
  emit ALSAOutputPortList(outputLastPorts);
  emit ALSAInputPortList(inputLastPorts);
}

/**
 * @short Directly connect track to client:port
 *
 * FIXME. It deletes ALL previous connections to track... maybe not
 * desired behaviour...
 */
void Midi::connectTrack(Track *tr, int client, int port){
  printf("Connect %s to %d:%d\n",tr->name(),client,port);

  // Mainly taken from ALSA documentation
  snd_seq_addr_t sender, dest;
  snd_seq_port_subscribe_t *subs;

  snd_seq_port_subscribe_alloca(&subs);
  sender.client = snd_seq_client_id(handle);
  sender.port = tr->getPortId();
  snd_seq_port_subscribe_set_sender(subs, &sender);

  ALSAPort *tport;
  for (tport=outputLastPorts.first();tport!=NULL;tport=outputLastPorts.next()){
    dest.client = tport->client;
    dest.port = tport->port;
    snd_seq_port_subscribe_set_dest(subs, &dest);
    snd_seq_unsubscribe_port(handle, subs);
  }

  
  dest.client = client;
  dest.port = port;
  snd_seq_port_subscribe_set_dest(subs, &dest);
  snd_seq_subscribe_port(handle, subs);
}

/**
 * @short Connects that output to the input of kmiditracker
 *
 */
void Midi::setInputPort(int nport){
  if (nport<0 || nport>=inputLastPorts.count())
    return;

  int client=inputLastPorts.at(nport)->client;
  int port=inputLastPorts.at(nport)->port;
  printf("Input set to %d:%d\n",client,port);

  // Mainly taken from ALSA documentation
  snd_seq_addr_t sender, dest;
  snd_seq_port_subscribe_t *subs;
  int ok;
  
  snd_seq_port_subscribe_alloca(&subs);

  sender.client = client;
  sender.port = port;
  snd_seq_port_subscribe_set_sender(subs, &sender);

  dest.client = snd_seq_client_id(handle);
  dest.port = currentPort;
  snd_seq_port_subscribe_set_dest(subs, &dest);

  ok=snd_seq_subscribe_port(handle, subs);
}
