/***************************************************************************
 *   Copyright (C) 2004,2005 by David Moreno Montero                       *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include "controller.h"
#include <qthread.h>

/**
 * @short Creates the controller with blank data
 *
 * Creates the arrays and copy initial data (size and midiout)
 */
Controller::Controller(MidiOut *out, unsigned int _size){
  size=_size;
  midi=out;
  minimum=0;
  maximum=127;
  bigstep=16;

  values=new float[size];
  use=new bool[size];

  unsigned int i;
  for (i=0;i<size;i++){
    values[i]=0.0;
    use[i]=false;
  }
  
  lock=new QMutex();
}

/**
 * @short Creates the Controller coping data from another Controller
 */
Controller::Controller(Controller *c){
  size=c->size;
  midi=c->midi;
  minimum=c->minimum;
  maximum=c->maximum;
  bigstep=c->bigstep;

  values=new float[size];
  use=new bool[size];

  unsigned int i;
  for (i=0;i<size;i++){
    values[i]=c->values[i];
    use[i]=c->use[i];
  }

  lock=new QMutex();
}

Controller::~Controller(){
  delete[] values;
  delete[] use;

  delete lock;
}

/**
 * @short Sets the controller value at the desired position
 *
 * If above or bellow there are passive values, change them too.
 */
void Controller::setController(unsigned int row, float val){
  lock->lock();
  if (row>=size){
    lock->unlock();
    return;
  }

  if (val<-1.0)
    val=-1.0;
  if (val>1.0)
    val=1.0;

  values[row]=val;
  use[row]=true;
  lock->unlock();
  emit controllerChanged(row, val);
  lock->lock();

  /* I have to change controllers above if not in real use, and below */
  int i,j,n;
  float ival,eval, value; /* initial value, end value */
  n=0;
  ival=eval=values[row];
  for (i=row-1;i>=0;i--){
    if (use[i]!=false){
      eval=values[i];
      break;
    }
    n++;
  }
  n++;
  j=0;
  /* simple linear interpolation */
  for (i=row-1,j=1;j<n;i--,j++){
    values[i]=((eval-ival)*(float(j)/n))+ival;
    value=values[i];
    lock->unlock();
    emit controllerChanged(i, value);
    lock->lock();
  }

  /* and now to down */
  n=0;
  ival=eval=values[row];
  for (i=row+1;i<(signed)size;i++){
    if (use[i]!=0){
      eval=values[i];
      break;
    }
    n++;
  }
  n++;
  j=0;
  /* simple linear interpolation */
  for (i=row+1,j=1;j<n;i++,j++){
    values[i]=((eval-ival)*(float(j)/n))+ival;
    value=values[i];
    lock->unlock();
    emit controllerChanged(i, value);
    lock->lock();
  }

  lock->unlock();
}

/**
 * @short Sets the controller value and wheter it is in use or not.
 */
void Controller::setController(unsigned int row, float val, bool inuse){
  lock->lock();
  if (row>size){
    lock->unlock();
    return;
  }
  lock->unlock();
  setController(row, val);
  lock->lock();
  use[row]=inuse;
  lock->unlock();
}

/**
 * @short Sets the controller value in use or not in use.
 */
void Controller::setControllerUse(unsigned int row,bool _use){
  lock->lock();
  if (row>size){
    lock->unlock();
    return;
  }
  use[row]=_use;

  /* Now i have to change the not used values */
  int i;
  float _value;
  for (i=row-1;i>=0;i--)
    if (use[i]){
      _value=values[i];
      lock->unlock();
      setController(i,_value);
      return;
    }
  /* o por arriba */
  for (i=row+1;i<(signed)size;i++)
    if (use[i]){
      _value=values[i];
      lock->unlock();
      setController(i,_value);
      return;
    }
  _value=values[row];
  lock->unlock();
  emit controllerChanged(row, _value);
}

/**
 * @short Changes the size of the Controller
 *
 * If the controller is smaller, just change the size parameter, but
 * do not reallocate arrays. It wastes memory, but saves time and
 * fragmentation. If bigger allocate a bigger array and copy data.
 */
void Controller::setSize(unsigned int nsize){
  lock->lock();
  if (nsize==size){
    lock->unlock();
    return;
  }
  else if (nsize<size){
    // FIXME, more complicated.. this way we save mem fragmentation?
    size=nsize;
  }
  else{
    float *nvalue=new float[nsize];
    bool *nuse=new bool[nsize];
    unsigned int i;
    for (i=0;i<size;i++){
      nvalue[i]=values[i];
      nuse[i]=use[i];
    }
    for (;i<nsize;i++){
      nvalue[i]=values[size-1];
      nuse[i]=false;
    }
    size=nsize;
    delete[] values; 
    delete[] use;
    values=nvalue;
    use=nuse;
  }
  lock->unlock();
}

/**
 * @short Sets the midi out.
 */
void Controller::setMidi(MidiOut *_midi){
  lock->lock();
  midi=_midi;
  lock->unlock();
}


float Controller::convertScale(float v){
  return (((v+1)/2.0)*(maximum-minimum)+minimum);
}
