/***************************************************************************
 *   Copyright (C) 2004,2005 by David Moreno Montero                       *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/


#ifndef __CONTROLLER_H__
#define __CONTROLLER_H__

#include <qobject.h>

class QMutex;
class MidiOut;

/**
 * @short Keeps the basic data for all controllers (midi, sysex...)
 *
 * The data is stored on an array of floats (usually 32 bits
 * precission). Values can be active or passive (use flag). Passive
 * values are interpolated from surrounding active values when
 * something changes.
 *
 * There is also a associated MidiOut, usually the same as the
 * parent's Pattern MidiOut.
 *
 * step(...) method control the real playing, but it is virtual and
 * should be reimplemented on the desired controllers.
 */
class Controller : public QObject{
  Q_OBJECT
protected:
  /// The array of values
  float *values;
  /// The active or passive flag
  bool *use;
  /// Real size of the controller array
  unsigned int size;

  /// Controls concurrency on this object
  QMutex *lock;
  /// Where to output midi data.
  MidiOut *midi;
  /// Minimum, maximum value, and bigstep
  int minimum, maximum, bigstep;
signals:
  /// The controller on this position have changed to value
  void controllerChanged(unsigned int row, float value);
  /// The controller text have changed
  void textChanged();
  /// Delete controller by name
  void delController(const QString &name);
  /// Deleted controller by name
  void deletedController(const QString &name);
  
public slots:
  void setController(unsigned int row, float val);
  void setController(unsigned int row, float val, bool inuse);
  void setControllerUse(unsigned int row, bool inuse);
  void setSize(unsigned int size);
  void setMidi(MidiOut *midi);

  void emitDeletedController(){ printf("dc\n"); emit deletedController(this->getText()); };

public:
  Controller(MidiOut *midi, unsigned int size=64);
  Controller(Controller *c);
  ~Controller();

  int getMinimum(){ return minimum; }
  int getMaximum(){ return maximum; }
  int getBigStep(){ return bigstep; }

  float convertScale(float v);
  float getOneStep(){ return 2.0/(maximum-minimum); };

  /// Should be reimplemented to output real midi data
  virtual void step(int step) = 0;
  /// Should be reimplemented with a type per controller type
  virtual int controllerType() = 0;
  /// Text to show on GUI things for this controller
  virtual QString getText() = 0;

  /// Gets the controller value where appropiate
  float getController(unsigned int row){ return (row<size) ? values[row] : -2.0; };
  /// Gets if the controller is in use
  bool getControllerUse(unsigned int row){ return (row<size) ? use[row] : false; };
};

#endif
