/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <qwmatrix.h> 
#include <qpainter.h>
#include <qcursor.h>
#include <qbrush.h>
#include <klocale.h>
#include <kpopupmenu.h>

#include "controller.h"
#include "controllercanvas.h"
#include "patternview.h"
#include "verticaltextitem.h"
#include "midicontroller.h"
#include "pitchcontroller.h"
#include "sysexcontroller.h"
#include "sysexbag.h"
#include "sysex.h"

/**
 * @short Creates a controller view from a selected controller on a pattern
 *
 * The data is got from the real Pattern, but sizes and such from the
 * PatternView.
 *
 * The created ControllerCanvas have red marks if controller is in use
 * (manually set) or grey if no in use (automatically set). It is
 * configurable with keyboard or with mouse.
 */
ControllerCanvas::ControllerCanvas(PatternView *_patternview,
				   int _controller,QWidget *parent, 
				   const char *name):
  QCanvasView(parent,name,0){
  int height;
  beatLength=_patternview->getBeatLength();

  setFocusPolicy(StrongFocus);
  keypos=0;

  controller=_patternview->getPattern()->getController(_controller);

  setCanvas(&canvas);
  setVScrollBarMode(AlwaysOff);
  setHScrollBarMode(AlwaysOff);

  cheight=_patternview->cellHeight();
  height=_patternview->height();

  setFrameStyle(Box+Plain);
  setLineWidth(1);

  canvas.resize(127,height);
  canvas.setBackgroundColor(_patternview->backgroundColor());

  controllerNameLabel=new 
    VerticalText(controller->getText(), &canvas);
  controllerNameLabel->move(15,0);
  controllerNameLabel->setZ(10);
  controllerNameLabel->show();

  patternView=_patternview;
  size=patternView->getPattern()->getRows();
  controllerId=_controller;

  keyposmark=new QCanvasRectangle(0,-cheight,127, cheight, &canvas);
  setBackgroundMode(PaletteHighlight);
  keyposmark->setBrush(QBrush(backgroundColor()));
  keyposmark->setPen(QPen(backgroundColor()));
  keyposmark->setZ(-2);
  keyposmark->show();

  connect(controller, SIGNAL(controllerChanged(unsigned int,float)), 
	  this, SLOT(controllerChangedSlot(unsigned int,float)));

  connect(controller, SIGNAL(textChanged()),
	  this, SLOT(controllerTextChangedSlot()));

  connect(controller, SIGNAL(deletedController(const QString &)),
	  patternView, SLOT(deletedController(const QString &)));

  drawAll();
}

/**
 * @short Calls the deleteData() method.
 */
ControllerCanvas::~ControllerCanvas(){
  delete controllerNameLabel;
  deleteData();
}

/**
 * @short Manages key events
 *
 * Movement is allowed in the canvas (up and down) and sides change
 * controller value, and propagates the new value to the
 * Pattern. Space cancels the use of a controller. Other keys as
 * Enter, Prev Page, Next Page... are also used.
 */
void ControllerCanvas::keyPressEvent(QKeyEvent *e){
  //  canvas.setChanged(keyposmark->boundingRect());
  double step=controller->getOneStep();
  double bigstep=controller->getBigStep()*step;

  switch(e->key()){
  case Key_Up:
    {
    if (keypos>0)
      keypos--;
    setPosition(keypos);
    }
    break;
  case Key_Down:
    {
    if (keypos<size-1)
      keypos++;
    setPosition(keypos);
    }
    break;
  case Key_Left:
    controller->setController(keypos, controller->getController(keypos)-step); 
    break;
  case Key_Right:
    controller->setController(keypos, controller->getController(keypos)+step);
    break;
  case Key_PageUp:
    controller->setController(keypos, controller->getController(keypos)-bigstep); 
    break;
  case Key_PageDown:
    controller->setController(keypos, controller->getController(keypos)+bigstep); 
    break;
  case Key_Enter: 
  case Key_Return: 
    patternView->setPosition(keypos,controllerId+patternView->getPattern()->getCols()); 
    patternView->setFocus();
    break;
  case Key_Backspace:
  case Key_Delete:
  case Key_Space:
    controller->setControllerUse(keypos, false);
    break;
  }
}

/**
 * @short Controls mouse movement to change controller values
 *
 * Only usefull when using leftbutton
 */
void ControllerCanvas::contentsMouseMoveEvent(QMouseEvent* e){
  e=e;

  if (lastPressedButton==LeftButton){
    setCursor(QCursor(SizeHorCursor));
    setFocus();
    
    controller->setController(keypos, (e->x()/64.0)-1.0); 
  
    controllerChangedSlot(controllerId, keypos, controller->getController(keypos)); 
  }

}

/**
 * @short Gets the event to pass it to its parent.
 *
 * Simple work-around to make when you use wheel mouse over a canvas,
 * change the pattern position. If I dont do this, the canvas moves
 * inside the pattern's scrollview (just a few pixels).
 */
void ControllerCanvas::viewportWheelEvent(QWheelEvent *e){
  ((QObject*)patternView)->event(e);
}


/**
 * @short Focus got so we have to change colors and such.
 */
void ControllerCanvas::focusInEvent(QFocusEvent *e){
  e=e;
  setPosition(keypos);
  setBackgroundMode(PaletteHighlight);
  keyposmark->setBrush(QBrush(backgroundColor()));
  keyposmark->setPen(QPen(backgroundColor()));
  canvas.setChanged(keyposmark->rect());
  canvas.update();
}

/**
 * @short Focus out, so we change again colors
 */
void ControllerCanvas::focusOutEvent(QFocusEvent *e){
  e=e;
  focusReallyOut();
}

/**
 * @short Focus is really out 
 *
 * It contrasts to when have focus, but real focus is in the PatternView
 * (normally because of keyboard movement of the cursor).
 */
void ControllerCanvas::focusReallyOut(){
  canvas.setChanged(keyposmark->rect());
  keyposmark->setY(-1*cheight);
  canvas.setChanged(keyposmark->rect());
  canvas.setChanged(controllerNameLabel->boundingRect());
  canvas.update();
}

/**
 * @short Mouse is pressed and controlled
 *
 * If right button is pressed, we should set to no use the controller.
 */
void ControllerCanvas::contentsMousePressEvent(QMouseEvent* e){
  e=e;

  if (!hasFocus())
    patternView->focusInController(controllerId, e->y()/cheight);
  setPosition(e->y()/cheight);

  if (e->button()==4){
    controller->setControllerUse(keypos, false);
  }
  else if (e->button()==2){
    showControllerChangeMenu();
    //    emit showControllerMenu();
  }
  lastPressedButton=e->button();

  setFocus();
}

/**
 * @short Shows a change controller menu
 *
 * In that menu all midicontrollers appear, the pitch bend, and sysex
 * messages (maybe more).
 *
 * You choose one, then the controller changes to that.
 */
void ControllerCanvas::showControllerChangeMenu(){
  int n;
  int firstmidicontroller, pitchbend, firstsysex,sysexcount;
  Pattern *pattern=patternView->getPattern();
  
  KPopupMenu popup(this);
  popup.insertTitle(i18n("Change Controller to..."));
  
  // I create the midi menu entries, saving first position
  KPopupMenu midim(this);
  KPopupMenu *midimm[4];
  int i,j,k=0;
  for (i=0;i<4;i++){
    midimm[i]=new KPopupMenu(this);
    n=midim.insertItem(QString("Controllers %1-%2").arg(i*32).arg(i*32+31),
		       midimm[i]);
  }
  firstmidicontroller=n-1;
  for (i=0;i<4;i++){
    for (j=0;j<32;j++,k++){
      n=midimm[i]->insertItem(QString("%1: %2").
			      arg(k).arg(MidiController::controllerName(k)));
    }
  }
  
  // Now I create the sysex controllers
  firstsysex=n-1;
  sysexcount=0;
  KPopupMenu sysexm(this);
  SysExBag *bag=pattern->getSysExBag();
  for (i=0;i<signed(bag->count());i++){
    SysEx *sysex=bag->at(i);
    for (j=0;j<signed(sysex->numParams());j++,sysexcount++)
      sysexm.insertItem(QString("%1:%2").
			arg(sysex->getName()).
			arg(sysex->getName(j)));
  }


  // First level menus
  popup.insertItem("MIDI Controller",&midim);
  pitchbend=popup.insertItem("Pitch Bend");
  popup.insertItem("SysEx Controller",&sysexm);
  
  // Exec
  i=popup.exec(QCursor::pos());
  
  // Guess what was clicked, and do it!
  Controller *c;
  if (i<=firstmidicontroller && i>firstmidicontroller-128){   // To midi something
    pattern->changeController(controllerId,
			      c=new MidiController(controller,
						   -i+firstmidicontroller));
    emit changedController(controllerId);
  }
  else if(i==pitchbend){ // To pitch bend
    pattern->changeController(controllerId,
			      c=new PitchController(controller));
    emit changedController(controllerId);
  }
  else if(i<=firstsysex && i>firstsysex-sysexcount){ // To sysex
    k=-i+firstsysex;
    for (i=0;i<signed(bag->count());i++){
      SysEx *sysex=bag->at(i);
      for (j=0;j<signed(sysex->numParams());j++,k--)
	if (k==0){
	  pattern->changeController(controllerId,
				    c=new SysExController(controller,sysex,j));
	  emit changedController(controllerId);
	}
    }
  }
  
  for (i=0;i<4;i++)
    delete midimm[i];
}

/**
 * @short Sets the position of this ControllerCanvas
 */
void ControllerCanvas::setPosition(int row){

  patternView->setRowPos(row);

  canvas.setChanged(keyposmark->rect());
  QRect r=controllerNameLabel->boundingRect();
  if (r.bottom()+cheight>=(keypos*cheight))
    canvas.setChanged(r);

  keypos=row;
  keyposmark->setY(keypos*cheight);

  canvas.setChanged(keyposmark->rect());
  if (r.bottom()+cheight>=(keypos*cheight))
    canvas.setChanged(r);
  canvas.update();
}


/**
 * @short Mouse is released
 *
 * The shape of the cursor changes, and we gain again the focus.
 */
void ControllerCanvas::contentsMouseReleaseEvent(QMouseEvent* e){
  e=e;
  setCursor(QCursor(ArrowCursor));
  setFocus();
}

/**
 * @short Deletes all related data, as marks...
 */
void ControllerCanvas::deleteData(){
  int i;

  if (NULL!=mark){
    for (i=0;i<size;i++)
      delete mark[i];
    delete[] mark;
    mark=NULL;
  }
  if (NULL!=markConnection){
    for (i=0;i<size+1;i++)
      delete markConnection[i];
    delete[] markConnection;
    markConnection=NULL;
  }
  if (NULL!=markNumber){
    for (i=0;i<size;i++)
      delete markNumber[i];
    delete[] markNumber;
    markNumber=NULL;
  }

}

/**
 * @short Delete the data, and draw it again
 */
void ControllerCanvas::redrawAll(){
  deleteData();

  drawAll();
}

/**
 * @short Draws all data according to the data in the Pattern.
 */
void ControllerCanvas::drawAll(){
  int i;
  int contr=0, lx,ly;
  QPen pen;
  QCanvasLine *line;
  QCanvasText *text;

  controllerNameLabel->setText(controller->getText());

  lx=int((controller->getController(0)*63.5)+64);
  ly=0;
  mark=new QCanvasLine*[size];
  markConnection=new QCanvasLine*[size+1];
  markNumber=new QCanvasText*[size];
  for (i=0;i<size;i++){
    contr=int((controller->getController(i)*63.5)+64);
    line=new QCanvasLine(&canvas);
    mark[i]=line;
    line->setPoints(contr,i*cheight+2, contr,2+int((i*cheight)+(cheight*0.75)));
    pen=SolidLine;
    pen.setWidth(5);
    pen.setCapStyle(RoundCap);
    if (controller->getControllerUse(i))
      pen.setColor(getColor());
    else
      pen.setColor(gray);
    line->setPen(pen);
    line->show();
    line->setZ(-1);
    // Draw connection lines
    line=new QCanvasLine(&canvas);
    markConnection[i]=line;
    line->setPoints(lx,ly+2, contr,2+int((i*cheight)+(cheight*0.37)));
    lx=contr;
    ly=(i*cheight)+int(cheight*0.37);
    pen=SolidLine;
    pen.setColor(getColor(true));
    line->setPen(pen);
    line->show();
    // Draw the text
    text=new QCanvasText
      (QString::number((int)controller->convertScale
		       (controller->getController(i))), 
       &canvas);
    text->setY(i*cheight+2);
    text->setZ(2);
    QFont font=text->font();
    font.setPointSize(10);
    text->setFont(font);
    text->show();
    markNumber[i]=text;
    markNumber[i]->setX(127-markNumber[i]->boundingRect().right()+
			markNumber[i]->boundingRect().left()-2);

    // Maybe I have to draw the background
    if ((i%beatLength)==0){
      QCanvasRectangle *rec=new QCanvasRectangle(0,i*cheight, 127, cheight, &canvas);
      setBackgroundMode(PaletteMid);
      rec->setBrush(QBrush(backgroundColor()));
      rec->setPen(QPen(backgroundColor()));
      rec->setZ(-3);
      rec->show();
      
    }

  }
  // Draw final connection line
  line=new QCanvasLine(&canvas);
  markConnection[i]=line;
  line->setPoints(lx,ly, contr,size*cheight);
  lx=contr;
  ly=(i*cheight)+int(cheight*0.37);
  pen=SolidLine;
  pen.setColor(getColor(true));
  line->setPen(pen);
  line->show();
}

/**
 * @short Controller value changed, so we redraw.
 *
 * The redraw changes the mark position, the connection lines, and the
 * number label.
 */
void ControllerCanvas::controllerChangedSlot(unsigned int controllerCol, 
					     unsigned int row, float val){
  if (controllerCol!=(unsigned)controllerId)
    return;

  controllerChangedSlot(row,val);
}

void ControllerCanvas::controllerChangedSlot(unsigned int row, float val){
  int contr=int((val*63.5)+64);

  mark[row]->setPoints(contr,row*cheight+2, contr,2+int((row*cheight)+(cheight*0.75)));
  canvas.setChanged(mark[row]->boundingRect());

  QPen pen=SolidLine;
  pen.setWidth(5);
  pen.setCapStyle(RoundCap);
  if (controller->getControllerUse(row))
    pen.setColor(getColor());
  else
    pen.setColor(gray);
  mark[row]->setPen(pen);

  QPoint start, end;
  
  if (row!=0)
    start=markConnection[row]->startPoint();
  else
    start=QPoint(contr,0);
  markConnection[row]->setPoints(start.x(),start.y(), 
				 contr,2+int((row*cheight)+(cheight*0.37)));
  canvas.setChanged(markConnection[row]->boundingRect());

  if (row!=(unsigned int)size-1)
    end=markConnection[row+1]->endPoint();
  else
    end=QPoint(contr,2+int(((row+1)*cheight)+(cheight*0.37)));
  markConnection[row+1]->setPoints( contr,2+int((row*cheight)+(cheight*0.37)) , 
				    end.x(),end.y());
  canvas.setChanged(markConnection[row+1]->boundingRect());

  markNumber[row]->setText(QString::number((int)controller->convertScale
					   (controller->getController(row)))); 

  markNumber[row]->setX(127-markNumber[row]->boundingRect().right()+
			markNumber[row]->boundingRect().left()-2);
  QRect r=controllerNameLabel->boundingRect();
  if (r.contains(7, row*cheight))
    canvas.setChanged(r);

  canvas.update();
}

/**
 * @short Controller id changed, we should update labels
 *
 * The new text is got from ControllerCanvas::controllerName(int n).
 */
void ControllerCanvas::controllerTextChangedSlot(unsigned int controllerCol){
  if (controllerCol!=(unsigned)controllerId)
    return;

  controllerTextChangedSlot();
}

void ControllerCanvas::controllerTextChangedSlot(){
  controllerNameLabel->setText(controller->getText());
  canvas.setAllChanged();
  canvas.update();
}

QColor ControllerCanvas::getColor(bool dark){
  switch(controller->controllerType()){
  case 1:
    return dark ? Qt::darkRed : Qt::red;
  case 2:
    return dark ? Qt::darkGreen : Qt::green;
  case 3:
    return dark ? Qt::darkYellow : Qt::yellow;
  default:
    return dark ? Qt::black : Qt::black;
  }
}


void ControllerCanvas::controllerChanged(unsigned int cn, Controller *c){
  if (cn!=(unsigned)controllerId)
    return;
  controllerChanged(c);
}

void ControllerCanvas::controllerChanged(Controller *c){
  printf("Controller changed\n");
  controller=c;

  connect(controller, SIGNAL(controllerChanged(unsigned int,float)), 
	  this, SLOT(controllerChangedSlot(unsigned int,float)));

  connect(controller, SIGNAL(textChanged()),
	  this, SLOT(controllerTextChangedSlot()));

  connect(controller, SIGNAL(deletedController(const QString &)),
	  patternView, SLOT(deletedController(const QString &)));

  
  redrawAll();
  canvas.setAllChanged();
  canvas.update();
}
