/***************************************************************************
 *   Copyright (C) 2003 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <iostream>
using namespace std;
#include <stdio.h>

#include <qxml.h>
#include <qdom.h>
#include <qfile.h>

#include "controller.h"
#include "midicontroller.h"
#include "sysexcontroller.h"
#include "pitchcontroller.h"
#include "sysex.h"
#include "kmtfile.h"
#include "tracker.h"
#include "track.h"
#include "kmiditrackerview.h"
#include "patternview.h"
#include "ksbfile.h"


/**
 * @short Prepare some global data needed to load/save
 *
 * The XML document have doctype KMidiTracker
 */
KMTFile::KMTFile(const QString &_filename, kmiditrackerView *_view){
  filename=_filename;
  view=_view;

  doc=new QDomDocument("KMidiTracker");
}

/**
 * @short Frees memory
 *
 * Free the QDomDocument
 */
KMTFile::~KMTFile(){
  delete doc;
}

/**
 * @short Saves the current song.
 *
 * It uses QDomDocument to save all data to an open file.  Tries to
 * store all relevant information.
 */
int KMTFile::save(){
  printf("save to %s\n",filename.ascii());

  QFile file(filename);
  if (!file.open(IO_WriteOnly))
    return -1;


  QPtrList<Track> trackList;
  trackList=view->getTrackList();
  QDomElement root=doc->createElement("kmtfile");
  QDomAttr at=doc->createAttribute("version");
  at.setValue("0.5.9");
  root.setAttributeNode(at);
  doc->appendChild(root);
  root.appendChild(elementSessionData());

  QDomElement el;

  KSBFile ksb(doc,view->getSysExBag());
  el=ksb.saveData("sysexbag");
  root.appendChild(el);

  Track *i;
  int id;
  for (i=trackList.first(),id=0;i;i=trackList.next(),id++){
    el=elementTrack(i,id);
    root.appendChild(el);
  }

  QTextStream ts(&file);
  ts<<doc->toString();

  file.close();
  return 0;
}

/**
 * @short Creates a QDomElement from a Track
 *
 * This data should be used to save songs. It stores all relevant
 * data.
 */
QDomElement KMTFile::elementTrack(Track *tr,int id){
  int i;

  QDomElement el=doc->createElement("track");
  QDomElement pattern;
  QDomElement sequence;
  QDomElement a;
  QDomAttr at;
  QDomText t;

  at=doc->createAttribute("id");
  at.setValue(QString::number(id));
  el.setAttributeNode(at);

  at=doc->createAttribute("program");
  at.setValue(QString::number(tr->getMIDIProgram()));
  el.setAttributeNode(at);

  if (tr->pattern(0)->getMIDIChannel()!=1){
    at=doc->createAttribute("midi_channel");
    at.setValue(QString::number(tr->pattern(0)->getMIDIChannel()));
    el.setAttributeNode(at);
  }

  if (!tr->getALSAOutput().isNull()){
    at=doc->createAttribute("alsa_output");
    at.setValue(tr->getALSAOutput());
    el.setAttributeNode(at);
  }

  a=doc->createElement("name");
  t=doc->createTextNode(tr->name());
  a.appendChild(t);
  el.appendChild(a);

  // initial sysex 
  QPtrList<SysEx> initialSysEx=tr->getInitialSysEx();
  if (initialSysEx.count()!=0){
    SysEx *i;
    for (i=initialSysEx.first();i;i=initialSysEx.next()){
      QDomElement iel=doc->createElement("initial_sysex");
      QDomText ielt=doc->createTextNode(i->getName());
      iel.appendChild(ielt);
      el.appendChild(iel);
    }
  }

  pattern=doc->createElement("patterns");

  at=doc->createAttribute("size");
  at.setValue(QString::number(tr->npatterns()));
  pattern.setAttributeNode(at);
  for (i=0;i<tr->npatterns();i++)
    pattern.appendChild(elementPattern(tr->pattern(i),i));

  el.appendChild(pattern);

  sequence=doc->createElement("sequence");

  at=doc->createAttribute("size");
  at.setValue(QString::number(tr->getSequenceSize()));
  sequence.setAttributeNode(at);

  at=doc->createAttribute("step");
  at.setValue(QString::number(tr->getStepSize()));
  sequence.setAttributeNode(at);

  for (i=0;i<tr->getSequenceSize();i++){
    a=doc->createElement("item");
    t=doc->createTextNode(QString::number(tr->getPatternIdInSequence(i)));

    a.appendChild(t);
    sequence.appendChild(a);
  }

  el.appendChild(sequence);
  return el;
}

/**
 * @short Creates a QDomElement from a selected Pattern
 *
 * This data should be used to save the song.  It stores all note
 * information, the pattern name, size...
 */
QDomElement KMTFile::elementPattern(Pattern *pt,int id){
  QDomElement el=doc->createElement("pattern");
  QDomAttr at;
  QDomElement e;
  QDomText t;

  at=doc->createAttribute("id");
  at.setValue(QString::number(id));
  el.setAttributeNode(at);

  at=doc->createAttribute("rows");
  at.setValue(QString::number(pt->getRows()));
  el.setAttributeNode(at);

  at=doc->createAttribute("cols");
  at.setValue(QString::number(pt->getCols()));
  el.setAttributeNode(at);

  at=doc->createAttribute("controllers");
  at.setValue(QString::number(pt->getControllers()));
  el.setAttributeNode(at);

  e=doc->createElement("name");
  t=doc->createTextNode(pt->name());
  e.appendChild(t);
  el.appendChild(e);


  unsigned int i,j;
  QDomElement c;
  MidiNote event;
  for (i=0;i<pt->getCols();i++){
    c=doc->createElement("column");
    if (pt->isColumnMuted(i)){
      at=doc->createAttribute("muted");
      at.setValue(QString::number(1));
      c.setAttributeNode(at);
    }

    for (j=0;j<pt->getRows();j++){
      e=doc->createElement("note");
      event=pt->getEvent(j,i);
      t=doc->createTextNode(QString::number(event.note()));

      if (event.velocity()!=100 && event.velocity()!=0){
	at=doc->createAttribute("velocity");
	at.setValue(QString::number(event.velocity()));
	e.setAttributeNode(at);
      }

      e.appendChild(t);
      c.appendChild(e);
    }
    el.appendChild(c);
  }
  float value;
  for (i=0;i<(unsigned)pt->getControllers();i++){
    c=doc->createElement("controller");
    Controller *ct=pt->getController(i);
    if (pt->isColumnMuted(pt->getCols()+i)){
      at=doc->createAttribute("muted");
      at.setValue(QString::number(1));
      c.setAttributeNode(at);
    }
    
    int type=ct->controllerType();
    switch(type){
    case 1:
      {
	MidiController *mc=(MidiController*)ct;
	at=doc->createAttribute("controller");
	at.setValue(QString::number(mc->getControllerId()));
	c.setAttributeNode(at);
      }
      break;
    case 2:
      {
	SysExController *sc=(SysExController*)ct;
	SysEx *sysex=sc->getSysEx();
	at=doc->createAttribute("sysex");
	at.setValue(sysex->getName());
	c.setAttributeNode(at);
	at=doc->createAttribute("sysex_param");
	at.setValue(QString::number(sc->getParamNumber()));
	c.setAttributeNode(at);
      }
      break;
    case 3: // Pitch bend do not need nothing more
      break;
    default:
      qWarning("Controller type not saveable yet\n");
      qWarning(" Data might be loaded as another controller type\n");
    }

    /// Tags are saved in reverse order... this way type is first
    /// (aestetically better)
    at=doc->createAttribute("type");
    at.setValue(QString::number(type));
    c.setAttributeNode(at);

    for (j=0;j<pt->getRows();j++){
      e=doc->createElement("value");
      value=ct->getController(j);
      t=doc->createTextNode(QString::number(value));

      at=doc->createAttribute("inuse");
      at.setValue(QString::number(ct->getControllerUse(j)));
      e.setAttributeNode(at);

      e.appendChild(t);
      c.appendChild(e);
    }
    el.appendChild(c);
  }

  return el;
}

/**
 * @short Reads the session data and creates a QDomElement
 *
 * This element should be used to save the data.
 */
QDomElement KMTFile::elementSessionData(){
  QDomElement el=doc->createElement("session");

  QDomElement a;
  QDomText b;
  QDomAttr at;

  Tracker *tracker=view->getTracker();

  a=doc->createElement("bpm");
   b=doc->createTextNode(QString::number(tracker->getBPM()));
   a.appendChild(b);
  el.appendChild(a);

  a=doc->createElement("loop");
  at=doc->createAttribute("begin");
  at.setValue(QString::number(tracker->getLoopBegin()));
  a.setAttributeNode(at);
  at=doc->createAttribute("end");
  at.setValue(QString::number(tracker->getLoopEnd()));
  a.setAttributeNode(at);
  at=doc->createAttribute("loop");
  at.setValue(QString::number(tracker->isLoopActive()));
  a.setAttributeNode(at);
  el.appendChild(a);


  a=doc->createElement("signature");
  at=doc->createAttribute("beat_length");
  at.setValue(QString::number(tracker->beatLength()));
  a.setAttributeNode(at);
  at=doc->createAttribute("beats_per_bar");
  at.setValue(QString::number(tracker->beatsPerBar()));
  a.setAttributeNode(at);
  el.appendChild(a);
  

  return el;
}


/**
 * @short Loads the filename file
 *
 * First of all tries to open the selected file and process it. If
 * succesfull, deletes the current song (deletes Track's), and creates
 * the new tracks from the elements. Also reads the tempo, selecion
 * and such information and it is set.
 */
int KMTFile::load(){
  cout <<"Load!"<<endl;

  // I read the document
  QFile file(filename);
  if (!file.open(IO_ReadOnly))
    return -1;
  cout <<"Open"<<endl;
  if (!doc->setContent(&file)){
    file.close();
    return -1;
  }
  file.close();

  // Check is what I want.
  QDomElement root=doc->documentElement();
  if (root.tagName()=="kmtfile"){
    printf("kmt");
    // Is what i want, I delete old.. if an error happends, sorry.
    view->clearTracks();
    
    cout <<"Seems ok to me "<<endl;
    
    QDomNode n=root.firstChild();
    while(!n.isNull()){
      QDomElement e = n.toElement();
      if(!e.isNull()){
	if(e.tagName()=="sysexbag"){
	  SysExBag *sysexbag=view->getSysExBag();
	  KSBFile ksb(doc,sysexbag);
	  sysexbag->clear();
	  ksb.loadData(e);
	}
	else if(e.tagName()=="session")
	  readSession(e);
	else if (e.tagName()=="track")
	  view->addTrack(readTrack(e));
	else 
	  cout <<"Unknown element"<<e.tagName()<<endl;
      }
      n=n.nextSibling();
    }
  }
  
  return 0;
}

/**
 * @short Reads the data associated to the session (BPM, loop)
 */
void KMTFile::readSession(const QDomElement &session){
  QDomNode n=session.firstChild();
  while(!n.isNull()){
    QDomElement e = n.toElement();
    if(!e.isNull()){
      if(e.tagName()=="bpm"){ /// BPM
	float bpm=e.text().toFloat();
	cout <<"Set BPM to "<<bpm<<endl;
	view->setBPM(bpm);
      }
      else if (e.tagName()=="loop"){
	bool loop=(e.attribute("loop","0").toInt()==1);
	int begin=e.attribute("begin","-1").toInt();
	int end=e.attribute("end","-1").toInt();

	view->getTracker()->setLoopRegion(begin,end, loop);
      }
      else if (e.tagName()=="signature"){
	int a=e.attribute("beat_length","4").toInt();
	int b=e.attribute("beats_per_bar","4").toInt();

	view->getTracker()->setSignature(a,b);
      }

      else 
	cout <<"Unknown element"<<e.tagName()<<endl;
    }
    n = n.nextSibling();
  }
}

/**
 * @short Reads a track from the track QDomElement
 *
 * Reads all the fields, and deletegates the read of the patterns to
 * KMTFile::readPattern(...);
 *
 * Returns a Track.
 */
Track *KMTFile::readTrack(const QDomElement &track){
  Track *ret=new Track(view->getSysExBag(), view,"Unknowntrack_loaded");
  
  bool havePattern=false;
  bool haveSequence=false;

  ret->setMIDIChannel(track.attribute("midi_channel","1").toInt());

  ret->setALSAOutput(track.attribute("alsa_output","-1:-1"));
  
  ret->setMIDIProgram(track.attribute("program","1").toInt());


  QDomNode n=track.firstChild();
  while(!n.isNull()){
    QDomElement e = n.toElement();
    if(!e.isNull()){
      if(e.tagName()=="name"){ /// BPM
	ret->setNewName(e.text());
      }
      else if (e.tagName()=="initial_sysex"){
	SysExBag *sysexbag=ret->getSysExBag();
	ret->addInitialSysEx(sysexbag->byName(e.text()));
      }
      else if (e.tagName()=="patterns"){
	QDomNode N=e.firstChild();
	while(!N.isNull()){
	  QDomElement E=N.toElement();
	  if (!E.isNull()){
	    if (E.tagName()=="pattern"){
	      Pattern *pattern;
	      pattern=readPattern(E);
	      ret->addPattern(pattern);
	      havePattern=true;
	    }
	  }
	  N=N.nextSibling();
	}
      }
      else if (e.tagName()=="sequence"){
	int item;
	QDomNode N=e.firstChild();
	item=0;
	while(!N.isNull()){
	  QDomElement E=N.toElement();
	  if (!E.isNull()){
	    if (E.tagName()=="item"){
	      haveSequence=true;
	      ret->setPatternInSequence(item, E.text().toInt());
	      item++;
	    }
	  }
	  N=N.nextSibling();
	}
      }
      else 
	cout <<"Unknown element"<<e.tagName()<<endl;
    }
    n=n.nextSibling();
  }

  // If the track had no patterns at all
  if (!havePattern){
    cout <<"No patterns, adding one empty"<<endl;
    ret->addPattern();
  }
  if (!haveSequence){
    ret->setPatternInSequence(0,0);
    ret->setPatternInSequence(1,0xFFFF);
  }
  return ret;
}

/**
 * @short Reads the pattern embedded in pattern QDomElement
 *
 * Reads all the fields and creates a Pattern.
 */
Pattern *KMTFile::readPattern(const QDomElement &pattern){
  Pattern *ret;
  ret=new Pattern(view->getSysExBag(), view,"UnknownPattern_loaded");
  QDomNode n=pattern.firstChild();
  
  unsigned int cols=pattern.attribute("cols","1").toInt();
  unsigned int rows=pattern.attribute("rows","16").toInt();

  unsigned int col=0,row, controller=0;

  ret->setRows(rows);
  while(ret->getCols()<cols)
    ret->addColumn();
  

  while(!n.isNull()){
    QDomElement e = n.toElement();
    if(!e.isNull()){
      if(e.tagName()=="name"){ /// BPM
	ret->setName(e.text());
      }
      else if (e.tagName()=="column"){
	bool muted=e.attribute("muted","0").toInt();
	printf("tryadd\n");
	ret->setColumnMuted(col,muted);
	printf("ok\n");
	
	QDomNode N=e.firstChild();
	row=0;
	while(!N.isNull()){
	  QDomElement E=N.toElement();
	  if (!E.isNull()){
	    if (E.tagName()=="note"){
	      ret->setNote(row,col, MidiNote(E.text().toInt(),
					     E.attribute("velocity","100").toInt()));
	      row++;
	    }
	  }
	  N=N.nextSibling();
	}
	col++;
      }
      else if (e.tagName()=="controller"){
	QDomNode N=e.firstChild();
	QDomElement E;
	Controller *ct;
	switch(e.attribute("type","-1").toInt()){
	case 1:
	  ct=new MidiController(e.attribute("controller","-1").toInt(),
				ret->getMidi(), ret->getRows());
	  break;
	case 2:
	  {
	    printf("SysEx\n");
	    SysEx *sysex;
	    sysex=view->getSysExBag()->byName(e.attribute("sysex",""));
	    if (NULL==sysex)
	      ct=new MidiController(0,ret->getMidi(), ret->getRows());
	    else
	      ct=new SysExController(sysex,e.attribute("sysex_param","0").toInt(),
				     ret->getMidi(), ret->getRows());
	  }
	  break;
	case 3:
	  ct=new PitchController(ret->getMidi(), ret->getRows());
	  break;
	default:
	  ct=new MidiController(0,ret->getMidi(), ret->getRows());
	}

	ret->addController(ct);
	
	bool muted=e.attribute("muted","0").toInt();
	ret->setColumnMuted(cols+controller,muted);
	
	row=0;
	while(!N.isNull()){
	  E=N.toElement();
	  if (!E.isNull()){
	    if (E.tagName()=="value"){
	      ct->setController(row, E.text().toFloat(), E.attribute("inuse","0").toInt());
	      row++;
	    }
	  }
	  N=N.nextSibling();
	}
	controller++;

      }
      else 
	cout <<"Unknown element"<<e.tagName()<<endl;
    }
    n=n.nextSibling();
  }

  cout <<"new pattern "<<ret->name()<<endl;

  return ret;
}
