/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/
#include <qthread.h>

#include "sysex.h"

#include "sysexcontroller.h"
#include "pitchcontroller.h"
#include "midicontroller.h"
#include "pattern.h"
#include "midievent.h"
#include "alsamidi.h"

/**
 * @short Creates a pattern with the desired size
 *
 * The pattern is the basic container of notes. They are arranged in
 * columns and rows. Each row is a tie step, and columns are used to
 * have polyphony. The patterns have also controllers, that are fully
 * configurable.
 */
Pattern::Pattern(SysExBag *bag, QObject *parent, 
		 const char *nname, int nrows, int ncols):QObject(parent, nname){
  unsigned int i;

  lock=new QMutex;
  rows=nrows;
  cols=ncols;
  event=new MidiNote*[cols];
  sysexbag=bag;

  for (i=0;i<cols;i++){
    event[i]=new MidiNote[rows];
    columnMuted.append(false);
  }
  lastNote=new MidiNote[cols];

  controller.setAutoDelete(true);
}

/**
 * @short Creates a copy of a pattern
 *
 * Just copies a pattern to another. 
 */
Pattern::Pattern(const Pattern &pt):QObject(pt.parent(),pt.name()){
  unsigned int i,j;

  // General data
  lock=new QMutex;
  rows=pt.rows;
  cols=pt.cols;
  event=new MidiNote*[cols];
  midi=pt.midi;
  sysexbag=pt.sysexbag;

  // Event copy
  for (i=0;i<cols;i++){
    event[i]=new MidiNote[rows];
    for (j=0;j<rows;j++)
      event[i][j]=pt.event[i][j];
    columnMuted.append(pt.columnMuted[i]);
  }
  lastNote=new MidiNote[cols];
  
  // Controllers copy
  controller.setAutoDelete(true);
  for (i=0;i<pt.getControllers();i++){
    Controller *ct, *nct=NULL;
    ct=pt.getController(i);
    switch(ct->controllerType()){
      case 1:
	nct=new MidiController(ct, ((MidiController*)ct)->getControllerId());
	break;
      case 2:
	nct=new SysExController(ct, ((SysExController*)ct)->getSysEx(),
				((SysExController*)ct)->getParamNumber());
	break;
      case 3:
	nct=new PitchController(ct);
	break;
    }
    if (nct!=NULL)
      addController(nct);
    setColumnMuted(cols+i,pt.columnMuted[cols+i]);
  }
}


/**
 * @short Removes unused data
 */
Pattern::~Pattern(){
  unsigned int i;

  //  lock->lock();

  delete lock;
  for (i=0;i<cols;i++)
    delete[] event[i];
  delete[] event;
  delete[] lastNote;
}

/**
 * @short Sets the note at the desired position
 */
void Pattern::setNote(unsigned int row, unsigned int col, const MidiNote &newNote){
  lock->lock();
  if ((row<rows) && (col<cols))
    event[col][row]=newNote;
  lock->unlock();
}

/**
 * @short Sets the position on the pattern
 */
void Pattern::setPosition(unsigned int npos){
  lock->lock();
  pos=npos;
  lock->unlock();
}

/**
 * @short Sets the midi object that this pattern uses.
 */
void Pattern::setMidi(Midi *nmidi, int nportId){
  lock->lock();

  midi.setMidi(nmidi);
  midi.setPort(nportId);

  lock->unlock();
}

/**
 * @short Adds a column to the pattern
 *
 * Recreate the pattern in a way that a new column is added.
 */
void Pattern::addColumn(){
  lock->lock();
  MidiNote **newEvent=new MidiNote*[cols+1];
  MidiNote *lastNoteT=new MidiNote[cols+1];

  columnMuted.insert(columnMuted.at(cols),false);

  unsigned int i;
  for (i=0;i<cols;i++){
    newEvent[i]=event[i];
    lastNoteT[i]=lastNote[i];
  }
  newEvent[i]=new MidiNote[rows];
  delete[] event;
  event=newEvent;
  delete[] lastNote;
  lastNote=lastNoteT;

  cols++;
  lock->unlock();
}


/**
 * @short Deletes a column from the pattern
 *
 * Recreate the pattern in a way that a column is deleted.
 */
void Pattern::delColumn(){
  lock->lock();

  if ((cols-1)<1){
    lock->unlock();
    return;
  }
  columnMuted.remove(columnMuted.at(cols-1));


  /* I should delete lastNote too ("realloc" it), but 2 bytes is not
     enought memory to free it */

  MidiNote **newEvent=new MidiNote*[cols-1];
  unsigned int i;
  for (i=0;i<cols-1;i++)
    newEvent[i]=event[i];

  delete[] event;
  event=newEvent;
  cols--;

  lock->unlock();
}


/**
 * @short Sets a new number of rows.
 *
 * Given a new number of rows, it fully recreate the pattern to meet
 * the new row size.
 */
void Pattern::setRows(unsigned int nrows){
  lock->lock();

  unsigned int i,j;
  MidiNote **newEvent=new MidiNote*[cols];

  for (i=0;i<cols;i++){
    newEvent[i]=new MidiNote[nrows];
    for (j=0;j<rows && j<nrows;j++)
      newEvent[i][j]=event[i][j];
    delete[] event[i];
  }
  delete[] event;
  event=newEvent;

  /// Set rows in controllers...
  Controller *c;
  for (c=controller.first();c;c=controller.next())
    c->setSize(nrows);

  rows=nrows;

  lock->unlock();
  emit patternChanged();
}

/**
 * @short Gets the event that is at the desired position. 
 *
 * On error it returns first of all events.
 */
MidiNote &Pattern::getEvent(unsigned int row, unsigned int col){
  MidiNote *note;

  lock->lock();
  if ((row<rows)&&(col<cols))
    note=&event[col][row];
  else
    note=&event[0][0];
  lock->unlock();

  return *note;
}


/**
 * @short Executes the events at step
 *
 * If a note is already playing, it first stops it, and the launches
 * the new one, and this for all columns.
 */
void Pattern::step(int step){
  unsigned int i;

  lock->lock();
  pos=step;

  if (pos>=rows){
    lock->unlock();
    return;
  }

  if (!midi.isValid()){
    lock->unlock();
    return;
  }
  
  for (i=0;i<cols;i++)
    playNoteNoLock(event[i][pos], i);

  Controller *c;
  for (c=controller.first();c;c=controller.next(),i++){
    if (columnMuted[i])
      continue;
    c->step(pos);
  }

  lock->unlock();
}

/**
 * @short Just plays a not in the given column. Dont use the lock.
 *
 * Having this in a separate function allows easy play with keyboard.
 * I do not have lock as is supposed to be called in a locked mutex.
 */
void Pattern::playNoteNoLock(const MidiNote &note, unsigned int i){
  if (columnMuted[i])
    return;
  if (note.note()!=MidiNote::NoNote){
    /// First stop sounding note
    //if (note.note()!=MidiNote::StopNote)
      midi.playNote(lastNote[i]);
    lastNote[i]=MidiNote(note.note(), 0);
    /// Send new note
    if (note.note()!=MidiNote::StopNote)
      midi.playNote(note);
  }
}

/**
 * @short Just plays a note in the given column. 
 *
 * Having this in a separate function allows easy play with keyboard.
 */
void Pattern::playNote(const MidiNote &note, unsigned int col){
  lock->lock();
  playNoteNoLock(note,col);
  lock->unlock();
}


/**
 * @short Returns the array of playing notes
 *
 * Be care as it returns the real array, and this is
 * error-prone. FIXME.
 */
MidiNote *Pattern::getPlayingNotes(){
  return lastNote;
}

/** 
 * @short Free the lastNote array
 *
 * It sets the actual playing notes to NoNote. Usually used when stop
 * is pressed.
 */
void Pattern::setNoPlayingNotes(){
  unsigned int i;
  for (i=0;i<cols;i++)
    lastNote[i]=MidiNote(MidiNote::NoNote, 0);
}


/**
 * @short Sets the playing notes from a vector. 
 *
 * It is used, in cujunction to Pattern::getPlayingNotes() to
 * correctly set the end of a sequence; when the sequence changes the
 * array is gathered and transmited to the next pattern, so it can sto
 * the notes if necesary.
 */
void Pattern::setPlayingNotes(const QValueList<MidiNote> &notes){
  unsigned int i;
  lock->lock();
  
  for (i=0;i<cols;i++)
    if (i<notes.count())
      lastNote[i]=notes[i];
    else
      lastNote[i]=MidiNote();
      
  lock->unlock();
}

/**
 * @short Changes the MIDI channel of this pattern to n
 */
void Pattern::setMIDIChannel(int n){
  lock->lock();
  midi.setChannel(n);
  lock->unlock();
}


/**
 * @short Add a new controller collumn with an unknown id
 *
 * The ID will be the number of controllers.
 */
void Pattern::addController(){
  addController(getControllers());
}

/**
 * @short Adds a new controller with the desired id
 */
void Pattern::addController(int id){
  printf("Created controller id %d, number %d\n",id, getControllers());

  addController(new MidiController(id,&midi,rows));
}

/**
 * @short Simply adds a given controller
 */
void Pattern::addController(Controller *c){

  lock->lock();
  controller.append(c);
  int id=controller.count()-1;

  emit insertedController(id);
  connect(c, SIGNAL(delController(const QString &)),
	  this, SLOT(delController(const QString &)));

  columnMuted.append(false);
  lock->unlock();
}

void Pattern::changeController(unsigned int cn, Controller *c){
  // CHECKME maybe a memory leak here.
  lock->lock();
  controller.replace(cn,c);
  emit controllerChanged(cn,c);
  lock->unlock();
}


/**
 * @short deletes The last controller
 */
void Pattern::delController(){
  int n;
  if ((n=getControllers())==0)
    return;
  delController(getControllers()-1);
}

/**
 * @short deletes the desired controllers
 *
 * Move all other controller in a highest column to the left.
 */
void Pattern::delController(unsigned int col){
  printf("delete controller %d\n",col);


  lock->lock();
  controller.setAutoDelete(false);
  Controller *c=controller.at(col);
  controller.remove(col);
  controller.setAutoDelete(true);
  columnMuted.pop_front();
  lock->unlock();
  c->emitDeletedController();
  delete c;
}

/**
 * @short deletes the desired controllers
 *
 * Move all other controller in a highest column to the left.
 */
void Pattern::delController(const QString &name){
  printf("Delete by name (sysex?) %s\n",name.ascii());
  lock->lock();

  for (unsigned int i=0;i<getControllers();i++)
    if (getController(i)->getText()==name){
      controller.setAutoDelete(false);
      Controller *c=controller.at(i);
      controller.remove(i);
      controller.setAutoDelete(true);
      columnMuted.pop_front();
      lock->unlock();
      c->emitDeletedController();
      delete c;
      return;
    }
  lock->unlock();
}

/**
 * @short Inserts a note position in the current column.
 *
 * All notes bellow this position are moved, and the last note is
 * lost.
 */
void Pattern::insertNote(unsigned int rpos, unsigned int cpos){
  lock->lock();
  if (rpos>=rows || cpos>=cols){
    lock->unlock();
    return;
  }

  unsigned int i;
  for (i=rows-1;i>rpos;i--){
    event[cpos][i]=event[cpos][i-1];
    lock->unlock();
    emit changeNoteSignal(i,cpos, event[cpos][i]);
    lock->lock();
  }
  event[cpos][i]=MidiNote();
  lock->unlock();

  emit changeNoteSignal(i,cpos, event[cpos][i]);
}
 
/**
 * @short Deletes selected note position, moving all notes bellow
 *
 * The actual note is lost.
 */
void Pattern::deleteNote(unsigned int rpos, unsigned int cpos){
  lock->lock();
  if (rpos>=rows || cpos>=cols){
    lock->unlock();
    return;
  }
  unsigned int i;
  for (i=rpos+1;i<rows;i++){
    event[cpos][i-1]=event[cpos][i];
    lock->unlock();
    emit changeNoteSignal(i-1,cpos, event[cpos][i-1]);
    lock->lock();
  }
  event[cpos][i-1]=MidiNote();
  lock->unlock();

  emit changeNoteSignal(i-1,cpos, event[cpos][i-1]);
}


/**
 * @short Set the selected column to muted
 */
void Pattern::setColumnMuted(int col, bool mute){
  lock->lock();
  columnMuted[col]=mute;
  
  if (col<cols && mute)
    if (lastNote[col].note()!=MidiNote::StopNote){
      midi.playNote(lastNote[col]);
      lastNote[col]=MidiNote(MidiNote::NoNote,0);
    }
  


  lock->unlock();
  //  emit columnMuted(col, mute);
}
