/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#ifndef __PATTERN_H__
#define __PATTERN_H__

#include <qobject.h>
#include <qvaluelist.h> 
#include <qptrlist.h>

#include "midiout.h"
#include "midievent.h"

class QMutex;
class MidiEvent;
class Midi;
class Controller;
class SysExBag;

/**
 * @short The pattern data
 *
 * This class stores all the data a pattern need, from the events,
 * size, step...
 *
 * It also contais step() method that sends a midi event if
 * appropriate.
 */

class Pattern:public QObject{
  Q_OBJECT
private:
  /// The list of events arranged int [col][row]
  MidiNote **event;
  /// List of last played notes
  MidiNote *lastNote;
  /// Columns size
  unsigned int cols;
  /// Whether set column is muted or not. It includes controllers
  QValueList<bool> columnMuted;

  QPtrList<Controller> controller;
  /// Rows size
  unsigned int rows;
  /// Position (row number)
  unsigned int pos;
  /// Thread Lock to keep the patterns thread safe
  QMutex *lock;
  /// Midi object to output to
  MidiOut midi;
  /// The SysEx bag
  SysExBag *sysexbag;
 
  void playNoteNoLock(const MidiNote &, unsigned int col);
public:
  Pattern(SysExBag *bag, QObject *parent,const char *name, int nrows=16, int ncols=1);
  Pattern(const Pattern &);
  ~Pattern();
  /// Number of rows
  unsigned int getRows() const { return rows; }
  /// Number of cols
  unsigned int getCols() const { return cols; }
  /// Actual position
  unsigned int getPos() const { return pos; }
  /// Actual midi channel on the pattern
  int getMIDIChannel() const { return midi.getChannel(); }
  MidiNote &getEvent(unsigned int row, unsigned int col);
  Controller *getController(unsigned int controllerCol) const { 
    return ((QPtrList<Controller>*)&controller)->at(controllerCol); }
  /// Number of controllers
  unsigned int getControllers() const { return controller.count(); };
  /// Gets sysexbag
  SysExBag *getSysExBag() const { return sysexbag; }
  /// Gets midi
  MidiOut *getMidi(){ return &midi; }

  MidiNote *getPlayingNotes();
  void setPlayingNotes(const QValueList<MidiNote> &notes);
  bool isColumnMuted(int col){ return columnMuted[col]; }
  void playNote(const MidiNote &, unsigned int col);


signals:
  /// The controller on this position have changed to value
  void controllerChanged(unsigned int controllerCol, unsigned int row, float value);
  /// The controller itself changed
  void controllerChanged(unsigned int cn, Controller *c);
  /// A new controller have been added
  void insertedController(int col);
  /// The note on selected position have changed to e
  void changeNoteSignal(unsigned int row, unsigned int col, const MidiNote &e);
  /// Controller deleted
  void deletedController(const QString &name);
  /// Pattern chnaged its state
  void patternChanged();


public slots:
  void step(int step);
  void setNote(unsigned int row, unsigned int col, const MidiNote &newNote);
  void setPosition(unsigned int npos);
  void setMidi(Midi *midi, int portId);
  void addController();
  void delController();
  void delController(unsigned int controller);
  void delController(const QString &name);
  void addController(int id);
  void addController(Controller *c);
  void changeController(unsigned int cn, Controller *c);
  void setMIDIChannel(int);
  void addColumn();
  void delColumn();
  void setRows(unsigned int rows);
  void insertNote(unsigned int rpos, unsigned int cpos);
  void deleteNote(unsigned int rpos, unsigned int cpos);
  void setNoPlayingNotes();
  void setColumnMuted(int col, bool mute=true);
};

#endif
