
/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <qscrollview.h>
#include <qhbox.h>
#include <qvbox.h>
#include <qlabel.h>
#include <qpushbutton.h>
#include <qclipboard.h>

#include <kinputdialog.h>
#include <kapplication.h>
#include <kconfig.h>

#include "midievent.h"
#include "patternview.h"
#include "patternprop.h"
#include "controllersproperties.h"
#include "pattern.h"
#include "trackview.h"
#include "controllercanvas.h"
#include "verticallabel.h"

/**
 * @short Create a View of a Pattern
 *
 * The view contains columns for the notes, and maybe some
 * ControllerView.
 *
 * All data is stored in the model Pattern.
 */
PatternView::PatternView(Pattern *_pattern, TrackView *trackview, 
			 QWidget *parent, const char *name):
  QScrollView(parent, name){

  pattern=_pattern;
  trackView=trackview;
  
  setFocusPolicy(StrongFocus);
  setMouseTracking(true);
  numberList.setAutoDelete(true);
  lastStepNumber=NULL;
  baseOctave=3;
  beatLength=trackview->getTrack()->getBeatLength();

  rows=cols=0;
  cpos=rpos=0;


  
  patternProp=new PatternProp(this, "pattern_properties");
  patternProp->init(pattern);

  controllersProp=new ControllersProperties(pattern,
					    this,"controller_properties");

  constructView();
  noteview[0][0]->setFocus();
  
  connect(patternProp,SIGNAL(setName(const QString &)), this, SLOT(setName(const QString &)));
  connect(patternProp,SIGNAL(setRows(unsigned int)), this, SLOT(setRows(unsigned int)));
  connect(patternProp,SIGNAL(setCols(unsigned int)), this, SLOT(setCols(unsigned int)));
  connect(patternProp,SIGNAL(showControllerProperties()), controllersProp, SLOT(show()));

  connect(controllersProp, SIGNAL(showSysExEditor()), trackView, SIGNAL(showSysExEditor()));
  connect(controllersProp, SIGNAL(addController(Controller *)), 
	  this, SLOT(addController(Controller *)));

  connect(pattern, SIGNAL(insertedController(int)), 
	  controllersProp, SLOT(insertedController(int)));
  connect(pattern, SIGNAL(changeNoteSignal(unsigned int,unsigned int, const MidiNote &)),
	  this, SLOT(changeNote(unsigned int,unsigned int, const MidiNote &)));
  

  connect(this, SIGNAL(setRowsSignal(unsigned int)), pattern, SLOT(setRows(unsigned int)));
  connect(this, SIGNAL(showTrackMenu()), trackView, SIGNAL(showTrackMenu()));

  connect(trackview->getTrack(), SIGNAL(newSignature(int,int)),
	  this, SLOT(setSignature(int,int)));

  reloadSettings();
}

/**
 * @short Remove unwanted data from the view
 *
 * Most of them labels.
 */
PatternView::~PatternView(){
  delete patternProp;
  delete controllersProp;

  deleteView();
}

/**
 * @short Reloads the settings from the config file.
 */
void PatternView::reloadSettings(){
  KConfig *config=kapp->config();
  
  config->setGroup("MIDI");
  midiThru=config->readEntry("MidiThru","0")=="true";
  
  config->setGroup("General");
  followSong=config->readEntry("FollowSong","false")=="true";
}

/**
 * @short Sets the working pointer to a position
 *
 * The position is first checked for vailidity, and may be on the note
 * part or in the controllers.
 */
void PatternView::setPosition(unsigned int row, unsigned int col){
  if (row>=rows || col>=cols+pattern->getControllers()){
    return;
  }
  removeSelection();
  if (rpos<rows && cpos<cols+pattern->getControllers()){
    if (cpos<cols){
      if ((rpos%beatLength)==0)
        noteview[cpos][rpos]->setBackgroundMode(PaletteMid );
      else
        noteview[cpos][rpos]->setBackgroundMode(PaletteBackground);
    }
  }
  else{
    cpos=0; rpos=0;
  }

  if (cpos>=cols)
    controller.at(cpos-cols)->focusReallyOut();
  
  if (row<rows){
    rpos=row;
    if (col<cols){
      cpos=col;
      noteview[cpos][rpos]->setBackgroundMode(hasFocus() ? PaletteHighlight : PaletteLight);
    }
    else if (col<cols+pattern->getControllers()){
      cpos=col;
      controller.at(cpos-cols)->setPosition(rpos);
    }
  }
  if (cpos<cols){
    ensureVisible(noteview[0][0]->width()*cpos, rpos*noteview[0][0]->height());
    ensureVisible(noteview[0][0]->width()*(cpos+2), rpos*noteview[0][0]->height());
  }
  else{
    ensureVisible(noteview[0][0]->width()*cpos + (cpos-cols)*127, rpos*noteview[0][0]->height());
    ensureVisible(noteview[0][0]->width()*(cpos+2) + (cpos-cols+2)*127, rpos*noteview[0][0]->height());
  }
  emit positionChanged(cpos,rpos);
}

/**
 * @short Sets the row position
 *
 * only change the row position, and ensure its visibility.
 */
void PatternView::setRowPos(unsigned int r){
  if (r<rows)
    rpos=r;
  ensureVisible(cols*noteview[0][0]->width() + (cpos-cols)*127, rpos*noteview[0][0]->height());
  ensureVisible(cols*noteview[0][0]->width() + (cpos-cols+1)*127, rpos*noteview[0][0]->height());
  emit positionChanged(cpos,rpos);
}

/**
 * @short Manage key events as new notes and such.
 *
 * Manages key events that may be new notes, movement, change to
 * controller mode...
 *
 * If focus is on the PatternView and cursor is over controllers
 * movement moves the cursor, but if you press enter the focus goes to
 * the ControllerView. Enter returns focus to PatternView.
 */
void PatternView::keyPressEvent(QKeyEvent *e){
  unsigned char keyboard_map[]={
    90,83,88,68,67,86,71,66,72,78,74,77,

    81,50,87,51,69,82,53,84,54,89,55,85,73
  };
  /*
    60, 65, 90, 83, 88, 67, 70, 86, 71, 66, 72, 78, 
    77, 75, 44, 76, 46, 81, 50, 87, 51, 69, 52, 82, 
    84, 54, 89, 55, 85, 73, 57, 79, 48, 80, 39
    };*/
  unsigned int i;
  bool haveNewNote=false;
  MidiNote newNote;

  //printf("%d, \n",e->key());

  // I check for position movement
  switch (e->key()){
  case Key_Left:    setPosition(rpos, cpos-1);    break;
  case Key_Up:      setPosition(rpos-1, cpos);    break;
  case Key_Right:   setPosition(rpos, cpos+1);    break;
  case Key_Down:    setPosition(rpos+1, cpos);    break;

  case Key_PageUp:    setPosition(rpos-8, cpos);    break;
  case Key_PageDown:  setPosition(rpos+8, cpos);    break;

  case Key_Home: setPosition(0,       cpos); break;
  case Key_End: setPosition(rows-1,cpos); break;

  case '.':   newNote=MidiNote(); haveNewNote=true; break;
  case Key_Backspace: newNote=MidiNote(MidiNote::StopNote, 100); haveNewNote=true; break;

  case Key_Insert: pattern->insertNote(rpos, cpos); break;
  case Key_Delete: pattern->deleteNote(rpos, cpos); break;

  case Key_Enter: 
  case Key_Return: 
    if (cpos>=cols && unsigned(cpos-cols)<pattern->getControllers())
      controller.at(cpos-cols)->setFocus();
    break;
  case '<':{
    MidiNote actualNote=pattern->getEvent(rpos, cpos);
    newNote=MidiNote(actualNote.note(), actualNote.velocity()-1);
    changeNote(rpos, cpos, newNote);
  }
    break;
  case '>':{
    MidiNote actualNote=pattern->getEvent(rpos, cpos);
    newNote=MidiNote(actualNote.note(), actualNote.velocity()+1);
    changeNote(rpos, cpos, newNote);
  }
    break;
  default:
    //    cout <<e->key()<<endl;
    ;
  }

  // Check if a notekey is pressed
  for (i=0;i<sizeof(keyboard_map);i++){
    // I have found the key to midikey map
    if ((unsigned int)e->key()==keyboard_map[i]){
      newNote=MidiNote(i+baseOctave*12, 100);
      haveNewNote=true;
      break;
    }
  }

  if (haveNewNote && cpos<cols){
    if (midiThru)
      pattern->playNote(newNote, cpos);
    changeNote(rpos, cpos, newNote);
    setPosition(rpos+1, cpos);
  }
}


/**
 * @short Change position or show menu
 *
 * Depending on the button you press, a new position should be
 * selected, or a context menu may appear.
 */
void PatternView::contentsMousePressEvent(QMouseEvent * e){
  setFocus();

  if (e->button()==RightButton){
    emit showTrackMenu();
  }

  else{ 
    int x,y;
    
    x=e->x()-numberList.at(0)->width(); 
    y=e->y()-QFontMetrics(QFont()).height();
    
    x/=noteview[0][0]->width();
    y/=noteview[0][0]->height();
    
    setPosition(y,x);
  }
}

/**
 * @short When releasing, if at diferent note position, then is selection
 */
void PatternView::contentsMouseMoveEvent(QMouseEvent * e){
  setFocus();
  
  if (e->button()==RightButton){
    return;
  }
  else{ 
    int x,y,t,i,j;
    x=e->x()-numberList.at(0)->width(); 
    y=e->y()-QFontMetrics(QFont()).height();
    ensureVisible(x,y,100,100);
    
    x/=noteview[0][0]->width();
    y/=noteview[0][0]->height();
  
    int x2,y2;
    x2=cpos; y2=rpos;
    if (x!=x2 || y!=y2){
      if (x>x2){
        t=x;        x=x2;        x2=t;
      }
      if (y>y2){
        t=y;        y=y2;        y2=t;
      }
      newSelection(x,y,x2,y2);
    }
  }
}

/**
 * @short Sets the new selected area
 */
void PatternView::newSelection(int x1, int y1, int x2, int y2){
  if (x1<0 || x2<0 || y1<0 || y2<0 || x1>x2 || x2>=cols || y1>y2 || y2>=rows)
    return;
  
  removeSelection();
  selection=QRect(x1,y1, x2-x1+1, y2-y1+1);
  int i,j;
  for (i=0;i<selection.width();i++)
    for (j=0;j<selection.height();j++){
      noteview[i+selection.x()][j+selection.y()]->setBackgroundMode(PaletteHighlight);
    }
}

/**
 * @short Removes the selected area
 */
void PatternView::removeSelection(){
  int i,j;
  for (i=0;i<selection.width();i++)
    for (j=0;j<selection.height();j++){
      int rpos=j+selection.y();
      int cpos=i+selection.x();
      if ((rpos%beatLength)==0)
        noteview[cpos][rpos]->setBackgroundMode(PaletteMid );
      else
        noteview[cpos][rpos]->setBackgroundMode(PaletteBackground);
    }
  selection=QRect(0,0,0,0);
}

/**
 * @short Sets the note on desired position to newNote
 */
void PatternView::changeNote(unsigned int row, unsigned int col, const MidiNote &newNote){
  if (row>=rows || col>=cols)
    return;

  noteview[col][row]->setText(noteText(newNote));

  pattern->setNote(row, col, newNote);
}

/**
 * @short It adds a new note on current position, and advances
 * position
 */
void PatternView::addMidiNote(int step, const MidiNote &newNote){
  if (step>=0)
    changeNote(step, cpos, newNote);
  else{
    changeNote(rpos, cpos, newNote);
    setPosition(rpos+1, cpos);
  }
}


/**
 * @short Recontructs all the view
 *
 * First of all it deletes current view, and a new one is created from
 * scratch.
 */
void PatternView::reconstructView(){
  deleteView();
  constructView();
}

/**
 * @short Deletes all the view
 *
 * Deletes all associated data to this view that is regenerated via
 * PatternView::constructView()
 */
void PatternView::deleteView(){
  unsigned int i;

  columns->hide();
  delete columns;
  for (i=0;i<cols;i++)
    delete[] noteview[i];
  delete[] noteview;
  numberList.setAutoDelete(false); // Data already has been deleted
  numberList.clear();
  numberList.setAutoDelete(true); 
  lastStepNumber=NULL;
  muteButtons.clear();
}

/**
 * @short Construct the view from scratch
 *
 * Gets all data from the Pattern to create view of the current data.
 */
void PatternView::constructView(){
  unsigned int i,j;
  QVBox *vbox;
  QWidget *space;
  QPushButton *button;

  rows=pattern->getRows();
  cols=pattern->getCols();

  columns=new QHBox(viewport(),"pattern_columns");
  addChild(columns);
  columns->hide();

  vbox=new QVBox(columns);
  new QWidget(vbox);
  patternName=new VerticalLabel(pattern->name(),vbox);
  connect(patternName, SIGNAL(doubleClicked()), this, SLOT(askNewName()));

  noteview=new QLabel**[cols];
  vbox=new QVBox(columns);
  space=new QWidget(vbox);
  space->setFixedHeight(QFontMetrics(QFont()).height());
  QLabel *num;
  for (i=0;i<rows;i++){
    QString text="<font size=-8>";
    text+=QString::number(i+1);
    text+="</font></tt>";

    num=new QLabel(text, vbox);
    if ((i%beatLength)==0)
      num->setBackgroundMode(PaletteMid);
    numberList.append(num);
  }

  for (j=0;j<cols;j++){
    vbox=new QVBox(columns);
    button=new QPushButton("Mute",vbox);
    button->setFixedHeight(QFontMetrics(QFont()).height());
    button->setToggleButton(true);
    button->setOn(pattern->isColumnMuted(j));
    muteButtons.append(button);
    connect(button, SIGNAL(clicked()), this, SLOT(recheckMuteStatus()));

    noteview[j]=new QLabel*[rows];
    for (i=0;i<rows;i++){
      noteview[j][i]=new QLabel(noteText(pattern->getEvent(i,j)), vbox);
      if ((i%beatLength)==0)
	noteview[j][i]->setBackgroundMode(PaletteMid);
    }
  }
  unsigned int controllers=pattern->getControllers();

  controller.clear();
  for (j=0;j<controllers;j++){
    vbox=new QVBox(columns);
    button=new QPushButton("Mute",vbox);
    button->setFixedHeight(QFontMetrics(QFont()).height());
    button->setToggleButton(true);
    button->setOn(pattern->isColumnMuted(pattern->getCols()+j));
    muteButtons.append(button);
    connect(button, SIGNAL(clicked()), this, SLOT(recheckMuteStatus()));

    ControllerCanvas *cv=new ControllerCanvas(this, j, vbox, "ControllerCanvas");
    cv->controllerTextChangedSlot();

    connect(pattern, SIGNAL(controllerChanged(unsigned int,unsigned int,float)), 
	    cv, SLOT(controllerChangedSlot(unsigned int, unsigned int, float)));
    connect(cv, SIGNAL(showControllerMenu()),
	    trackView, SIGNAL(showTrackMenu()));
    connect(pattern, SIGNAL(controllerChanged(unsigned int, Controller *)),
	    cv, SLOT(controllerChanged(unsigned int, Controller *)));
    connect(cv, SIGNAL(changedController(unsigned int)),
	    controllersProp, SLOT(changedController(unsigned int)));
    
    cv->focusReallyOut();
    controller.append(cv);
  }


  columns->show();

  setPosition(rpos,cpos);
}

/**
 * @short Returns the correct text of a MidiNote
 *
 * The text returned is used to be shown on the PatternView.
 */
QString PatternView::noteText(const MidiNote &newNote){
  QString note[]={
    "C-", "C#",
    "D-", "D#",
    "E-", 
    "F-", "F#",
    "G-", "G#",
    "A-", "A#",
    "B-"
  };

  // If I have a new note somewhere, place it correctly
  QString notetext="<tt>";
  if (newNote.note()==MidiNote::NoNote){
    notetext+="--- <font size=-4>---</font></tt>";
  }
  else if (newNote.note()==MidiNote::StopNote){
    notetext+="###";
    notetext+=" <font size=-4>";
    notetext+=QString::number(newNote.velocity());
    notetext+="</font></tt>";
  }
  else{
    notetext+=note[newNote.note()%12];
    notetext+=QString::number(newNote.note()/12);
    notetext+=" <font size=-4>";
    notetext+=QString::number(newNote.velocity());
    notetext+="</font></tt>";
  }
  return notetext;
}


/**
 * @short Adds a column to the pattern, and recontructs the view
 */
void PatternView::addColumn(){
  pattern->addColumn();
  reconstructView();
}

/**
 * @short Deletes a column from the pattern, and recontructs the view
 */
void PatternView::delColumn(){
  pattern->delColumn();
  reconstructView();
}

/**
 * @short Adds a controller from the pattern, and recontructs the view
 */
void PatternView::addController(){
  pattern->addController();
  reconstructView();
}

/**
 * @short Adds a given controller from the pattern, and recontructs the view
 */
void PatternView::addController(Controller *c){
  pattern->addController(c);
  reconstructView();
}

/**
 * @short Deletes a controller from the pattern, and recontructs the view
 */
void PatternView::delController(){
  pattern->delController();
}

/**
 * @short Deletes a selected controller from the pattern, and recontructs the view
 */
void PatternView::delController(int controller){
  pattern->delController(controller);
}
/**
 * @short Deletes a selected controller from the pattern, and recontructs the view
 */
void PatternView::delController(const QString &name){
  printf("Delete controller by name\n");
  pattern->delController(name);
}

void PatternView::deletedController(const QString &){
  reconstructView();
}


/**
 * @short Sets the number of rows of the Pattern
 */
void PatternView::setRows(unsigned int nrows){
  if (nrows==rows)
    return;
  emit setRowsSignal(nrows);
  reconstructView();
}

/**
 * @short Sets the number of columns 
 * 
 * Adds or removes patterns to get the desired number
 */
void PatternView::setCols(unsigned int ncols){
  if (ncols==cols)
    return;

  // Maybe its not very pretty, but works as expected.
  while (pattern->getCols()>ncols)
    pattern->delColumn();
  while (pattern->getCols()<ncols)
    pattern->addColumn();
  reconstructView();
}

/**
 * @short Sets the name of the Pattern, and changes references
 */
void PatternView::setName(const QString &name){
  pattern->setName(name);
  patternName->setText(name);
  patternProp->init(pattern);
  trackView->setPatternName(patternId, name);
}

/**
 * @short Sets the base octave for the keyboard input
 */
void PatternView::setBaseOctave(int octave){
  baseOctave=octave;
}

/**
 * @short Sets the pattern id
 */
void PatternView::setPatternId(int id){
  patternId=id;
}

/**
 * @short Change status to reflect focus in
 *
 * Colors are changed to reflect where is the focus. It checks if we
 * are over a controller to set the color on it, and re-sets position
 */
void PatternView::focusInEvent(QFocusEvent *){
  unsigned int i;
  for (i=0;i<pattern->getControllers();i++)
    if ((i+cols)!=cpos)
      controller.at(i)->focusReallyOut();
  
  setPosition(rpos,cpos);
}

/**
 * @short Change status to reflect focus out
 *
 * Re-set position to change it
 */
void PatternView::focusOutEvent(QFocusEvent *){
  setPosition(rpos,cpos);
}

/**
 * @short Focus is over a controller, change colors
 */
void PatternView::focusInController(int n, int row){
  if ((rpos<rows) && (cpos<cols)){
    if ((rpos%beatLength)==0)
      noteview[cpos][rpos]->setBackgroundMode(PaletteMid);
    else
      noteview[cpos][rpos]->setBackgroundMode(PaletteBackground);
  }
  unsigned int i;
  for (i=0;i<pattern->getControllers();i++)
    if (i!=(unsigned)n)
      controller.at(i)->focusReallyOut();
  cpos=cols+n;
  rpos=row;
}

/**
 * @short Clears the step position unsetting colors on the numbers
 */
void PatternView::clearStepPosition(){
  if (lastStepNumber!=NULL)
    lastStepNumber->unsetPalette();
}

/**
 * @short Redraws step position. 
 *
 * Maybe it changes the lastStepNumber color, and changes the actual
 * position color.
 */
void PatternView::drawStepPosition(){
  QLabel *l=numberList.at(pattern->getPos());
  if (l!=NULL){
    l->setPaletteBackgroundColor(Qt::yellow);
    if (followSong)
      ensureVisible(contentsX()+50,noteview[0][0]->height()*pattern->getPos());
  }
  lastStepNumber=l;
}

/**
 * @short Gets the Pattern name
 */
QString PatternView::getPatternName(){
  return pattern->name();
}

/**
 * @short Height of all the view in pixels
 *
 * It is number of rows x row height
 */
int PatternView::height(){
  return (pattern->getRows()*cellHeight());
}

/**
 * @short Height of a single cell (all cells are the same in height)
 */
int PatternView::cellHeight(){
  return (noteview[0][0]->minimumSizeHint().height());
}

/**
 * @short Show the properties box.
 */
void PatternView::showProperties(void){
  patternProp->show();
}

/**
 * @short Show the controller properties box.
 */
void PatternView::showControllersProperties(void){
  controllersProp->show();
}

/**
 * @short Checks the mute buttons list to check witch ones are muted or not
 *
 * With that data it tells the real pattern to mute them
 */
void PatternView::recheckMuteStatus(){
  int i;
  QPushButton *I;
  for (I=muteButtons.first(),i=0;I;I=muteButtons.next(),i++)
    if (I->isOn())
      pattern->setColumnMuted(i,true); 
   else
      pattern->setColumnMuted(i,false);
}

/**
 * @short Asks for a new name
 */
void PatternView::askNewName(){
  QString name;
  bool ok;

  name=KInputDialog::getText("Pattern Name","New Pattern Name",pattern->name(),&ok,this);
  if (ok)
    setName(name);
}


/**
 * @short Copies selected data to the clipboard
 */
void PatternView::doCopy(){
  QString str,s;
    
  unsigned int i,j;
  for (j=selection.y();j<selection.y()+selection.height();j++){
    for (i=selection.x();i<selection.x()+selection.width();i++){
      s=noteview[i][j]->text();
      s.replace("<tt>","");
      s.replace("<font size=-4>","");
      s.replace("</font></tt>","");
      str+=s+"\t";
    }
    str+="\n";
  }
  
  QClipboard *cb = QApplication::clipboard();
  cb->setText(str);
}

/**
 * @short Pastes from clipborad to the current position
 */
void PatternView::doPaste(){
  QString note[]={
    "C-", "C#",
    "D-", "D#",
    "E-", 
    "F-", "F#",
    "G-", "G#",
    "A-", "A#",
    "B-"
  };

  QClipboard *cb = QApplication::clipboard();
  QStringList sl=QStringList::split("\n",cb->text());
  QStringList::Iterator i,endI;
  int k,noten,velocity;
  MidiNote newNote;
  for(i=sl.begin(),endI=sl.end();i!=endI;++i){ // I go throught the rows
    QStringList sl2=QStringList::split("\t",*i);
    QStringList::Iterator j,endJ;
    int J;
    for(j=sl2.begin(),endJ=sl2.end(),J=0;j!=endJ;j++,J++){ // and inside throught the cols
      QString s;
      noten=-1;
      // guess what note is it
      s=(*j).left(2);
      if (s=="--")
        newNote=MidiNote(MidiNote::NoNote,0);
      else if (s=="##"){
        velocity=(*j).mid(4,3).toInt();
        newNote=MidiNote(MidiNote::StopNote,velocity);
      }
      else{
        for (k=0;k<12;k++){
          if (s==note[k]){
            noten=k;
            break;
          }
        }
        // if it is a note
        if (noten!=-1){
          // add the octave
          noten+=(*j).mid(2,1).toInt()*12;
          // and finally velocity
          velocity=(*j).mid(4,3).toInt();
          
          newNote=MidiNote(noten,velocity);
        }
      }
      changeNote(rpos, cpos+J, newNote);
    }
    if (rpos+1>=rows){ // break if no space left down, and already written
      setPosition(rpos+1, cpos);
      return;
    }
    setPosition(rpos+1, cpos);
  }
}

