/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <qvbox.h>
#include <ctype.h>
#include <limits.h>
#include <qvariant.h> 

#include "sequenceview.h"
#include "track.h"

/**
 * @short Creates the sequence view
 *
 * Creates the basic view that consists of several columns of
 * labels. The first one have the numbers (in step increments), the
 * seccond and following the tracks arrangments.
 *
 * You can choose if there should be titles or not in the view
 */
SequenceView::SequenceView(bool drawTitles, QWidget *parent, const char *name)
  :QScrollView(parent,name){

  setFocusPolicy(StrongFocus);
  hbox=new QHBox(viewport(), "hbox_sequenceview");
  addChild(hbox);
  step=16;
  size=0;
  cpos=0; rpos=0;
  lastNumberLabel=NULL;

  unsigned int i;
  QVBox *vbox=new QVBox(hbox);
  QLabel *nlabel;
  if (drawTitles)
    new QLabel(QString(""),vbox);
  for (i=0;i<size;i++){
    nlabel=new QLabel(QString("<tt><font size=-8>%1</font></tt>").arg((i*step)+1,-3), vbox);
    nlabel->setBackgroundMode(PaletteMid);
    numberList.append(nlabel);
  }
  column.append(vbox);
  column.setAutoDelete(true);
  label.setAutoDelete(true);
  numberList.setAutoDelete(true);
  loopBegin=-1; loopEnd=-1; loopFirstClick=-1;

  setPosition(0,0);
  showTitles=drawTitles;
}

/**
 * @short Deletes unwanted data
 *
 * Everything is autodeleted
 */
SequenceView::~SequenceView(){
} 

/**
 * @short add a track to manage
 *
 * The added track is managed. You can manage several tracks if you
 * want. The labels are created and if the size need to be
 * incremented, it is.
 */
void SequenceView::manageTrack(Track *tr){
  unsigned int j,i;

  if (tr==NULL)
    return;

  if (size<unsigned(tr->getSequenceSize())){
    setSizeSlot(tr->getSequenceSize());
  }
  else{
    tr->setPatternInSequence(size-1,tr->getPatternIdInSequence(size-1));
  }

  trackList.append(tr);

  QVBox *qv;
  qv=new QVBox(hbox,"track_sequence_column");
  column.append(qv);

  if (showTitles){
    QLabel *title;
    title=new QLabel(QString("<center>%1</center>").arg(tr->name()),qv);
    title->setBackgroundMode(PaletteMid);
    trackLabelList.append(title);
  }
  // I add a column
  QPtrList<QLabel> *clabel=new QPtrList<QLabel>;
  clabel->setAutoDelete(true);
  i=label.count();
  // and all labels
  for (j=0;j<size;j++){
    QLabel *l=new QLabel(textOf(j*step, i),qv);
    l->setBackgroundMode(PaletteBackground);
    clabel->append(l);
  }

  label.append(clabel);


  connect(tr, SIGNAL(patternChanged()), this, SLOT(redrawPatterns()));
  connect(tr, SIGNAL(sequenceSizeChanged(unsigned int)),this, SLOT(setSizeSlot(unsigned int)));
  connect(tr, SIGNAL(redrawStepPointer(int)), this, SLOT(setPosition(int)));
  connect(tr, SIGNAL(deleteTrack(Track*)), this, SLOT(unmanageTrack(Track*)));
  if (showTitles){
    connect(tr, SIGNAL(nameChanged(Track *,const QString &)), this, SLOT(trackNameChanged(Track *,const QString&)));
    connect(tr, SIGNAL(muteChanged(Track *, bool)), this, SLOT(muteChanged(Track *,bool)));
  }
  qv->show();
}

/**
 * @short Unmanages selected track (mostly because of deletion).
 *
 * It removes all data refering to this track. Signals and such are
 * deleted by Qt.
 */
void SequenceView::unmanageTrack(Track *tr){
  int id;

  // First I find out the column number and track id
  id=trackList.findRef(tr);
  if (id==-1)    return;
  
  // Then I delete the labels pointer for this row, and move the
  // followings one position to the left.
  label.remove(id);

  // I delete the column, remove it, and the track from the list
  column.remove(id+1);
  trackList.remove(id);
  if (showTitles)
    trackLabelList.remove(id);
 
}


/**
 * @short Text of the pattern at that sequence position.
 *
 * Returns the text that shoud be at a desired position acording to
 * the patterns that lies in it.
 */
QString SequenceView::textOf(int row, int trackN){
  if (trackN<signed(trackList.count())){
    Track *track=trackList.at(trackN);
    QString name;
    bool first;
    if (track!=NULL){
      name=track->patternNameAtRow(row,first);
      if (first)
	return QString("<tt>%1</tt>").arg(name.left(20),-20);
      else
	return QString("<center> <font size=-4 color=grey><tt>%1</tt></font>&nbsp;</center>").
	  arg(name.left(20),-20);
    }
  }
  return QString("<tt>====================</tt>");
}

/**
 * @short Slot used to increment the size of the sequence view.
 *
 * Increments the size adding needed labels and numbers.
 *
 * Note that the size can also be reduced.
 */
void SequenceView::setSizeSlot(unsigned int n){
  if (size==n)
    return;
  // Grow
  else if (n>size){
    // First the number
    QVBox *vbox=column.at(0);
    QLabel *nlabel;
    unsigned int i,j;
    for (i=size;i<n;i++){
      nlabel=new QLabel(QString("<tt><font size=-8>%1</font></tt>").arg((i*step)+1,-3), vbox);
      nlabel->setBackgroundMode(PaletteMid);
      numberList.append(nlabel);
      nlabel->show();
    }
    // Then the labels
    QPtrList<QLabel> *clabel;
    QLabel *l;
    for (i=0;i<label.count();i++){
      vbox=column.at(i+1);
      clabel=label.at(i);
      for (j=size;j<n;j++){
	l=new QLabel(textOf(j*step,i),vbox);
	l->setBackgroundMode(PaletteBackground);
	l->show();
	clabel->append(l);
      }
    }
  }
  // Shrink
  else{
    // First I check if it is allowable
    Track *t;
    for (t=trackList.first();t;t=trackList.next())
      if (unsigned(t->getSequenceSize())>n)
        return;

    if (rpos>=n)
      setPosition(rpos-1,cpos);
    unsigned int i,j;
    for (i=n;i<size;i++){
      if (numberList.at(numberList.count()-1)==lastNumberLabel)
         lastNumberLabel=NULL;
      numberList.remove(numberList.count()-1);
    }
    // Then the labels
    QPtrList<QLabel> *clabel;
    for (i=0;i<label.count();i++){
      clabel=label.at(i);
      for (j=n;j<size;j++)
	      clabel->remove(clabel->count()-1);
    }
  }
  show();
  size=n;
  emit sizeChanged(n);
}

/**
 * @short gives focus to this widget
 *
 * 
 */
void SequenceView::contentsMousePressEvent(QMouseEvent *ev){
  setFocus();

  if (numberList.count()<=0)
    return;
  if (ev->button()==RightButton){
    loopFirstClick=-1;
    loopBegin=-1; loopEnd=-1;
    emit newLoop(loopBegin*step, (loopEnd+1)*step-1);
    return;
  }

  if (ev->x()<numberList.at(0)->width()){
    int n=ev->y()/numberList.at(0)->height();
    if (showTitles) n--;
    loopFirstClick=n;
  }
  else{
    if (trackList.count()<=0)
      return;
    int x,y;
    x=(ev->x()-numberList.at(0)->width())/label.at(0)->at(0)->width();
    y=ev->y()/label.at(0)->at(0)->height();
    if (showTitles){
      y--;
      if (y<0){
	if (x<signed(trackList.count()))
	  if (trackList.at(x)->isMutted())
	    trackList.at(x)->unmute();
	  else
	    trackList.at(x)->mute();
      }
    }

    setPosition(y,x);
  }
}

/**
 * @short Check of mouse movement
 *
 * If the movement have been on the numbers, the a new region is going
 * to be selected.
 */
void SequenceView::contentsMouseMoveEvent(QMouseEvent *ev){
  if (numberList.count()<=0)
    return;
  // Only events when on the numbers
  if (loopFirstClick==-1)
    return;

  int n=ev->y()/numberList.at(0)->height();
  if (showTitles) n--;

  if (n>loopFirstClick){
    loopBegin=loopFirstClick;
    loopEnd=n;
  }
  else{
    loopBegin=n;
    loopEnd=loopFirstClick;
  }

  redrawNumbers();

  if (column.count()>1)
    ensureVisible(0, 
		  (0.5+n+(showTitles ? 1 : 0))*label.at(0)->at(0)->height());
}

/**
 * @short The new loop region is selected
 *
 * The new region is selected... if apropiate.
 */
void SequenceView::contentsMouseReleaseEvent(QMouseEvent *ev){
  if (ev->button()==LeftButton && loopFirstClick!=-1){
    emit newLoop(loopBegin*step, (loopEnd+1)*step-1);
  }
  loopFirstClick=-1;
}


/**
 * @short Receive keyboard events and process them
 *
 * Translates keyboard to movement on the list and/or set of new
 * patterns. Actually a maximun of 10 patterns can be added beacuse of
 * this funcion. FIXME.
 */
void SequenceView::keyPressEvent(QKeyEvent *ev){
  bool ok=false;
  switch(ev->key()){
  case Key_Left:
    setPosition(rpos, cpos-1);
    ok=true;
    break;
  case Key_Right:
    setPosition(rpos, cpos+1);
    ok=true;
    break;
  case Key_Up:
    setPosition(rpos-1, cpos);
    ok=true;
    break;
  case Key_Down:
    setPosition(rpos+1, cpos);
    ok=true;
    break;
  case Key_Enter:
  case Key_Return:
    if (cpos<trackList.count())
      emit setActivePatternViewInSequence(trackList.at(cpos)->getPatternIdInSequence(rpos));
    ok=true;
    break;
  case Key_Backspace:
    trackList.at(cpos)->setPatternInSequence(rpos, 0xFFFF);
    label.at(cpos)->at(rpos)->setText(textOf(rpos*step,cpos));
    setPosition(rpos+1, cpos);
    ok=true;
    break;
  case Key_Insert:
    trackList.at(cpos)->insertPatternInSequence(rpos);
    ok=true;
    break;
  case Key_Delete:
    trackList.at(cpos)->removePatternInSequence(rpos);
    ok=true;
    break;
  }

  if (ok==false && isalnum(ev->key())){
    int npattern=-1;
    if (isdigit(ev->key()))
      npattern=ev->key()-'0';
    if (isalpha(ev->key()))
      npattern=ev->key()-'A'+10;

    trackList.at(cpos)->setPatternInSequence(rpos, npattern);
    label.at(cpos)->at(rpos)->setText(textOf(rpos*step,cpos));
    setPosition(rpos+1, cpos);
  }
}

/**
 * @short Sets the position on the view
 *
 * If the position is the last position, a new position is added and
 * cursor is set to there.
 */
void SequenceView::setPosition(unsigned int _rpos, unsigned int _cpos){
  if (_rpos<size && _cpos<(column.count()-1)){
    label.at(cpos)->at(rpos)->setBackgroundMode(PaletteBackground);
    rpos=_rpos; cpos=_cpos;
    label.at(cpos)->at(rpos)->setBackgroundMode(hasFocus() ? PaletteHighlight : PaletteLight);

  }
  else if (_rpos==size){
    unsigned int i;
    for (i=0;i<trackList.count();i++)
      trackList.at(i)->setPatternInSequence(_rpos, 
					    trackList.at(i)->getPatternIdInSequence(_rpos));
    /// Size must have changed because of signals/slots
    if (_rpos<size)
      setPosition(_rpos,_cpos);
  }
  if (column.count()>1){
    ensureVisible((cpos+1)*label.at(0)->at(0)->width(), 
		  (rpos+(showTitles ? 1 : 0))*label.at(0)->at(0)->height());
    ensureVisible((cpos)*label.at(0)->at(0)->width(), 
		  (1+rpos+(showTitles ? 1 : 0))*label.at(0)->at(0)->height());
  }
  emit positionChanged(cpos,rpos);
}

/**
 * @short Redraw the patterns
 *
 * It re-sets the labels of the patterns names. Used when a pattern
 * name changes, and this things.
 */
void SequenceView::redrawPatterns(){
  unsigned int i,j;
  
  for (i=0;i<column.count()-1;i++)
    for (j=0;j<size;j++){
      label.at(i)->at(j)->setText(textOf(j*step,i));
    }
  setPosition(rpos,cpos);
}

/**
 * @short Redraw numbers acording to loop
 *
 * Redraws the numbers of the left column to reflect the loop.
 */
void SequenceView::redrawNumbers(){
  int i;
  for (i=0;i<signed(size);i++){
    if (i>=loopBegin && i<=loopEnd)
      numberList.at(i)->setBackgroundMode(PaletteHighlight);
    else
      numberList.at(i)->setBackgroundMode(PaletteMid);
  }
  if (NULL!=lastNumberLabel)
    lastNumberLabel->setPaletteBackgroundColor(Qt::yellow);
  
}


/**
 * @short The widget gets the focus
 *
 * So it needs to redraw the cursor to reflect the correct color.
 */
void SequenceView::focusInEvent(QFocusEvent *){
  setPosition(rpos,cpos);
}
/**
 * @short The widget loses focus
 *
 * The position is set again to draw the cursor in other color.
 */
void SequenceView::focusOutEvent(QFocusEvent *){
  setPosition(rpos,cpos);
}

/**
 * @short Sets the position of the step pointer. 
 *
 * The pointer is only drawn at the row numbers, and the color is
 * fixed. Maybe this should be configurable.
 */
void SequenceView::setPosition(int s){
  s/=step;
  if (unsigned(s)>size)
    return;
  QLabel *l=numberList.at(s);
  if (lastNumberLabel!=NULL){
    lastNumberLabel->unsetPalette();
    int i=numberList.findRef(lastNumberLabel);
    if (i>=loopBegin && i<=loopEnd)
      lastNumberLabel->setBackgroundMode(PaletteHighlight);
    else
      lastNumberLabel->setBackgroundMode(PaletteMid);
  }
  if (NULL!=l)
    l->setPaletteBackgroundColor(Qt::yellow);
  lastNumberLabel=l;
}

/**
 * @short Select new loop
 *
 * A new loop is set, and I must set some variables and redraw
 * numbers to reflect it.
 */
void SequenceView::loopChanged(int begin, int end){
  loopBegin=begin/step;
  loopEnd=((end+1)/step)-1;
  redrawNumbers();
}

/**
 * @short The name of the track changed.
 *
 * As the name of the track changed, if you have titles on this
 * sequenceview, they should be updated and drawn correctly.
 */
void SequenceView::trackNameChanged(Track *tr,const QString &newName){
  if (!showTitles)
    return;

  int id;
  id=trackList.findRef(tr);
  
  QLabel *label;
  label=trackLabelList.at(id);
  if (NULL==label)
    return;
  if (tr->isMutted())
    label->setText(QString("<center>(%1)</center>").arg(newName));
  else
    label->setText(QString("<center>%1</center>").arg(newName));
}

/**
 * @short Change name to refflect mute status
 */

void SequenceView::muteChanged(Track *tr, bool mute){
  if (!showTitles)
    return;
  int id;
  id=trackList.findRef(tr);
  QLabel *label;
  label=trackLabelList.at(id);
  if (NULL==label)
    return;
  if (mute)
    label->setText(QString("<center>(%1)</center>").arg(tr->name()));
  else
    label->setText(QString("<center>%1</center>").arg(tr->name()));

}

void SequenceView::setSignature(int a,int b){
  step=a*b;

  QLabel *i;
  int n;
  for (i=numberList.first(),n=0;i;i=numberList.next(),n++){
    i->setText(QString("<tt><font size=-8>%1</font></tt>").arg(n*step+1));
  }
}
