/***************************************************************************
 *   Copyright (C) 2003 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <iostream>
using namespace std;
#include <stdio.h>

#include <qfile.h>

#include "controller.h"
#include "midicontroller.h"
#include "sysexcontroller.h"
#include "pitchcontroller.h"
#include "sysex.h"
#include "smffile.h"
#include "tracker.h"
#include "track.h"
#include "kmiditrackerview.h"
#include "patternview.h"
#include "ksbfile.h"


SMFFile::SMFFile(const QString &_filename, kmiditrackerView *_view){
  view=_view;
  filename=_filename;
}

SMFFile::~SMFFile(){

}

/// This should always work, even on Big Endian machines. Sorry for the overhead there
static long int long_2be(int x){
  union{
    char c[4];
    int r;
  };
  r=0;
  c[0]=(x>>24)&0x0FF;
  c[1]=(x>>16)&0x0FF;
  c[2]=(x>>8)&0x0FF;
  c[3]=x&0x0FF;

  return r;
}

/// This should always work, even on Big Endian machines. Sorry for the overhead there
static short int short_2be(short int x){
  union{
    char c[2];
    short int r;
  };
  r=0;
  c[0]=(x>>8)&0x0FF;
  c[1]=x&0x0FF;

  return r;
}

int SMFFile::save(){
  struct header_t
  {
    char           id[4];  /* This will be 'M','T','h','d' */
    unsigned long int len; /* This will be 6 */
    
    unsigned short int fmt;
    unsigned short int nTracks;
    unsigned short int signature;
  };
  QPtrList<Track> tl=view->getTrackList();
  struct header_t header={ {'M','T','h','d'},
			   long_2be(6), short_2be(1),
			   short_2be(tl.count()+1), // One extra track for initial tempo
			   short_2be(4)
  };

  QFile file(filename);
  if (!file.open(IO_WriteOnly))
    return -1;

  file.writeBlock((char*)&header,14);
  
  // Tempo at start
  {
    file.writeBlock("MTrk",4);
    unsigned long int len=long_2be(7);
    file.writeBlock((char*)&len,4);
    file.putch(0x00);
    file.putch(0xFF);
    file.putch(0x51);
    file.putch(0x03);


    int l=int(((60.0e6/view->getTracker()->beatLength())/view->getTracker()->getBPM())*4);

    file.putch((l&0x00FF0000)>>16);
    file.putch((l&0x0000FF00)>>8);
    file.putch((l&0x000000FF));
  }

  Track *t;
  for (t=tl.first();t;t=tl.next())
    saveTrack(t,file);


  file.close();

  return 0;
}

/**
 * @short Saves a tracker
 * 
 * A track have a initial state (name, program, channel), and then several patterns
 */
void SMFFile::saveTrack(Track *t, QFile &file){
  file.writeBlock("MTrk",4);
  QValueList<char> data;
  unsigned int i;

  // Now I must compose the message

  // track name
  data.append(0x00); // time 0
  data.append(0xFF);
  data.append(0x03);
  QString name=t->name();
  data.append(name.length());
  const char *c=name.ascii();
  for (i=0;i<name.length();i++){
    data.append(c[i]);
  }

  // Instrument
  data.append(0x00);
  data.append(0xC0+t->getMIDIChannel()-1);
  data.append(t->getMIDIProgram());

  // Real track data
  step=0; 
  memset(lastNote,0,256);
  Pattern *pt=NULL,*tpt;
  int pstep=0;
  //printf("%d patterns\n",t->getSequenceSize());
  for (i=0;i<(unsigned)t->getSequenceSize();i++){
    tpt=t->pattern(t->getPatternIdInSequence(i));
    //printf("%d %08X\n",i,tpt);
    if (tpt!=NULL){
      pt=tpt;
      pstep=0;

      saveTime(data,step);
      data.append(0x0FF);
      data.append(0x01);
      int j,l=strlen(pt->name());
      data.append(l);
      for (j=0;j<l;j++)
        data.append(pt->name()[j]);
      step=0;
    }
    savePattern(pt,data,pstep);
    pstep+=16;
  }

  // And the last notes
  unsigned int j;
  for (j=0;j<256;j++)
    if (lastNote[j]!=0){
      saveTime(data,step);
      data.append(0x90+t->getMIDIChannel()-1);
      data.append(lastNote[j]);
      data.append(0);
      step=0;
    }
  
  // Finally the write of the data
  unsigned long int len=long_2be(data.count());
  file.writeBlock((char*)&len,4);

  for (i=0;i<data.count();i++)
    file.putch(data[i]);
}

/**
 * @short Saves a simple pattern
 */
void SMFFile::savePattern(Pattern *p, QValueList<char> &data, int pstep){
  unsigned char lastType=0x00;
  unsigned int i,j;

//  printf("save pattern %s\n",p->name());

  if (NULL==p){
    return;
  }
  unsigned char channel=(p->getMIDIChannel()-1)&0x0F;
  for (i=0;i<16;i++){
    if (i+pstep<p->getRows()){
      // Notes
      for (j=0;j<p->getCols();j++){
        MidiNote ev=p->getEvent(i+pstep,j);
        if (!p->isColumnMuted(j))
          if (ev.note()!=MidiNote::NoNote){
            saveTime(data,step);
            if (lastType!=0x90){
              data.append(0x90+channel);
              lastType=0x90;
            }
            if (lastNote[j]!=0){
              data.append(lastNote[j]);
              data.append(0);
              data.append(0);
            }
            data.append(ev.note());
            data.append(ev.velocity());
            step=0;
            lastNote[j]=ev.note();
          }
      }
      // Controllers
      for (j=0;j<p->getControllers();j++){
        if (!p->isColumnMuted(j+p->getCols())){
          Controller *c=p->getController(j);
          switch(c->controllerType()){
            case 1: // Midi Controller
              saveTime(data,step);
              step=0;
              if (lastType!=0xB0){
                data.append(0xB0+channel);
                lastType=0xB0;
              }
              data.append(((MidiController*)c)->getControllerId());
              data.append((unsigned char)c->convertScale(c->getController(i+pstep)));
              break;
            case 3: // pitch bend
              saveTime(data,step);
              step=0;
              if (lastType!=0xE0){
                data.append(0xE0+channel);
                lastType=0xE0;
              }
              {
                int v=((int)c->convertScale(c->getController(i+pstep)))+0x2000;
                data.append((unsigned char)(v&0x7F));
                data.append((unsigned char)((v>>7)&0x7F));
              }
              break;
            default:
              fprintf(stderr,"This kind of controller is unkwnon to save to .mid (%d).\n",c->controllerType());
          }
        }
      }
    }
    step+=1;
  }
}


/**
 * @short Saves delta-time
 *
 * Code from http://www.borg.com/~jglatt/tech/midifile.htm
 */
void SMFFile::saveTime(QValueList<char> &data, unsigned long value){
  unsigned long buffer;
   buffer = value & 0x7F;

   while ( (value >>= 7) ){
     buffer <<= 8;
     buffer |= ((value & 0x7F) | 0x80);
   }

   while (1){
      data.append(buffer&0x0FF);
      if (buffer & 0x80)
          buffer >>= 8;
      else
          break;
   }
}

int SMFFile::load(){
  printf("load midi\n");

  return -1;
}
