/***************************************************************************
 *   Copyright (C) 2004,2005 by David Moreno Montero                       *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <assert.h>

#include "midiout.h"
#include "sysex.h"
#include "sysexcontroller.h"

/**
 * @short Creates a new SysEx controller 
 *
 * From the given sysex message and param number, to the said MidiOut
 * and with the given size.
 */
SysExController::SysExController(SysEx *_sysex, int nparm, 
				 MidiOut *midi, unsigned int size):Controller(midi,size){
  sysex=_sysex;
  paramn=nparm;
  switch(_sysex->paramLength(paramn)){
  case 1:
    break;
  case 2:
    maximum=127*127;
    bigstep=bigstep*bigstep;
    break;
  }
  connect(_sysex, SIGNAL(deletedParam(const QString &, int)),
	  this, SLOT(delSysExParam(const QString &,int)));
}

/**
 * @short Copys a controller to convert it to a SysExController
 */
SysExController::SysExController(Controller *c, SysEx *_sysex, int _nparam)
  :Controller(c){
  sysex=_sysex;
  paramn=_nparam;
  switch(_sysex->paramLength(paramn)){
  case 1:
    break;
  case 2:
    maximum=127*127;
    bigstep=bigstep*bigstep;
    break;
  }
}

/**
 * @short Performs a step of the SysEx controller
 */
void SysExController::step(int step){
  if (step>=1 && step<=signed(size)){
    float value=values[step-1];
    if (value!=lastvalue || step==1){
      sysex->setParam(paramn, value);

      QByteArray msg=sysex->getSysEx();
      midi->sendSysEx(msg.data(),msg.size());


      int i=0;
      while((msg[i]&0x0FF)!=0x0F7){
	printf("%02X ",msg[i]&0x0FF);
	i++;
      }
      printf("%02X\n",msg[i]&0x0FF);
      
      lastvalue=value;
    }
  }
}

/**
 * @short Shown text of the controller
 */
QString SysExController::getText(){
  return QString("%1 - %2").arg(sysex->getName()).arg(sysex->getName(paramn));
}


/**
 * @short Changes the sysex associated with this controller.
 *
 * The new sysex is given as well as the new parameter number
 */
void SysExController::changeSysEx(SysEx *nsysex, int nparam){
  sysex=nsysex;
  paramn=nparam;

  emit textChanged();
}

/**
 * @short This slot is received when a Sysex is deleted
 *
 * If first checks if the sysex is the associated one, and if so
 * deletes the controller.
 */
void SysExController::delSysExParam(const QString &name, int _paramn){
  if (name==sysex->getName() && _paramn==paramn){
    printf("Del sysexparam SysExController::...\n");
    emit delController(getText());
  }
}
