/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                       *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <qmutex.h>
#include <qstring.h>

#include "track.h"
#include "pattern.h"
#include "alsamidi.h"
#include "midievent.h"
#include "sysexbag.h"
#include "sysex.h"

/**
 * @short Creates a new empty Track
 *
 * Creates anew empty track, with one Pattern and a basic
 * sequence. Basic step size is 16.
 */
Track::Track(SysExBag *_sysexbag, QObject *parent, const char *nname):QObject(parent,nname){
  lock=new QMutex;
  midi=NULL;
  sysexbag=_sysexbag;
  portId=0;
  stepSize=16;
  beatLength=4;
  firstStepInCurrentPattern=0;

  currentPattern=NULL;
  MIDIChannel=1;
  MIDIProgram=0;
  muteFlag=false;
  patternList.setAutoDelete(true);
  ALSAoutput="";
}

/**
 * @short Deletes correctly the data (the lock)
 */
Track::~Track(){
  emitDeleteTrack(this);
  midi->delPort(portId);
  delete lock;
}

/**
 * @short Resturn a pattern by list id, or current if -1
 *
 * Returns a pattern by it's number on the list. If the number is -1,
 * it returns current pattern.
 */
Pattern *Track::pattern(int n){
  Pattern *ret;

  lock->lock();
  if (n==-1)
    ret=currentPattern;
  ret=patternList.at(n);
  lock->unlock();

  return ret;
}

/**
 * @short plays a selected step of the track
 *
 * It plays the step note on the selected pattern. If the step is a
 * multiply of the step size maybe a pattern change is needed acording
 * to sequence, so it looks this condition, copy the playing notes to
 * a temporal buffer and transmit this list to the following
 * pattern. Finally sets the current pattern the new selected pattern.
 */
bool Track::step(int step){
  bool stop=false;
  lock->lock();

  if (muteFlag){
    lock->unlock();
    return stop;
  }


  if ((step%stepSize)==0){
    if (sequence.at(step/stepSize)!=NULL){
      unsigned int i;
      firstStepInCurrentPattern=step;
      
      if (currentPattern!=NULL){
	MidiNote *notes=currentPattern->getPlayingNotes();
	for (i=currentPattern->getCols();i--;)
	  if (!lastNotesPlayed.empty())
	    lastNotesPlayed.pop_front();
	for (i=currentPattern->getCols();i--;){
	  lastNotesPlayed.push_front(notes[i]);
	}
      }

      currentPattern=sequence.at(step/stepSize);
      if (NULL!=currentPattern)
	currentPattern->setPlayingNotes(lastNotesPlayed);
    }
    else{ /// Just an optimization, at the end of sequence there is no
	 /// pattern, and the check is only needed on these cases
      if (unsigned(step)>=(sequence.count()*stepSize)){
	stop=true;
      }
    }
  }
  
  if (NULL!=currentPattern)
    currentPattern->step(step-firstStepInCurrentPattern);
  
  lock->unlock();
  return stop;
}

/**
 * @short Stops all playign notes on this track
 *
 * All actual playing notes are retreived, and then the stop event is
 *sent. The lastNotesPlayed list is also cleared.
 */
void Track::stopNotes(){
  // First I get actually playing notes
  if (currentPattern!=NULL){
    unsigned int i;
    MidiNote *notes=currentPattern->getPlayingNotes();
    for (i=currentPattern->getCols();i--;)
      if (!lastNotesPlayed.empty())
	lastNotesPlayed.pop_front();
    for (i=currentPattern->getCols();i--;){
      lastNotesPlayed.push_front(notes[i]);
    }
    currentPattern->setNoPlayingNotes();
  }
  // Now I stop them
  QValueList<MidiNote>::iterator note;
  for (note=lastNotesPlayed.begin();note!=lastNotesPlayed.end();++note)
    midi->playNote(MidiNote((*note).note(),0),MIDIChannel,portId);
  lastNotesPlayed.clear();
}

/**
 * @short Sets the Midi object on this track
 * 
 * The Midi object is changed here and on all the patterns it
 * contains.
 */
void Track::setMidi(Midi *_midi){
  lock->lock();

  midi=_midi;
  portId=midi->addPort(name());

  if (ALSAoutput!=""){
    printf("New output %s\n",ALSAoutput.ascii());
    int cid, pid,ok;
    
    ok=sscanf(ALSAoutput.ascii(),"%d:%d",&cid, &pid);
    printf("Selected %d %d\n",cid,pid);
    if (ok==2)
      midi->connectTrack(this,  cid, pid);
  }
  else{
    printf("No output defined\n");
  }

  Pattern *pattern;
  for (pattern=patternList.first();pattern!=NULL;pattern=patternList.next())
    pattern->setMidi(midi,portId);

  lock->unlock();
}

/**
 * @short Add a new empty Pattern to the track
 *
 * An empty pattern is added.
 */
void Track::addPattern(void){
  lock->lock();

  Pattern *pattern;

  pattern=new Pattern(sysexbag,this, 
		      QString(name())+" "+QString::number(patternList.count()),
		      stepSize);

  lock->unlock();

  addPattern(pattern);
}



/**
 * @short Add a new Pattern to the track
 *
 * The passed pattern is now managed by this track
 */
void Track::addPattern(Pattern *pattern){
  if (pattern==NULL)
    return;

  lock->lock();

  pattern->setMidi(midi, portId);
  patternList.append(pattern);
  pattern->setMIDIChannel(MIDIChannel);

  lock->unlock();

  connect(this,SIGNAL(setMIDIChannelSignal(int)), pattern, SLOT(setMIDIChannel(int)));
  connect(pattern,SIGNAL(patternChanged()), this, SIGNAL(patternChanged()));
  emit newPattern(pattern);
}



/**
 * @short Add a new Pattern to the track
 *
 * The passed pattern is now managed by this track
 */

void Track::delPattern(Pattern *pattern){
  printf("trying to del pattern\n");
  if (pattern==NULL)
    return;
  printf("del pattern!!! %s\n",pattern->name());

  lock->lock();
  unsigned int i;
  for (i=0;i<sequence.count();i++)
    if (sequence.at(i)==pattern)
      sequence.replace(i,NULL);
  if (currentPattern==pattern)
    currentPattern=NULL;
  patternList.setAutoDelete(false);
  patternList.remove(pattern);
  patternList.setAutoDelete(true);
  lock->unlock();

  emit deletedPattern(pattern);
  emit patternChanged();
  delete pattern;
}


/**
 * @short Returns the name of the pattern that is at row
 *
 * If there is no pattern at this row, it returns a fixed QString.
 *
 * In first it returns, if applicable, whether the returned pattern is
 * the first of the sequence list if the name is larger than a step.
 *
 * First also says when out of bounds
 */
QString Track::patternNameAtRow(int row, bool &first){
  QString ret;
  first=true;
  lock->lock();
  int pos=row/stepSize;
  if (sequence.at(pos)!=NULL)
    ret=sequence.at(pos)->name();
  else{
    if (unsigned(pos)>=sequence.count())
      first=false;
    ret="--------------------";
    int i=pos-1;
    while(i>=0){
      if (sequence.at(i)!=NULL)
	if ( signed((sequence.at(i)->getRows()+stepSize-1)/stepSize) > signed(pos-i)){
	  first=false;
	  ret=sequence.at(i)->name();
	}
      i--;
    }
  }
  lock->unlock();
  
  return ret;
}

/// Returns the size of the track (number of sequences)
int Track::getSequenceSize(){
  return sequence.count();
}

/// Returns the step size (step distance between sequences)
int Track::getStepSize(){
  return stepSize;
}

/**
 * @short Set the new track a desired sequence position
 *
 * If the sequence position does not exist, a new sequence position is
 * added. Just one position is added independently of the patternList
 * size.
 * 
 * patternChanged() signal is emited.
 */
void Track::setPatternInSequence(unsigned int rpos, unsigned int npattern){
  lock->lock();
  if (rpos<sequence.count()){
    sequence.remove(rpos);
    sequence.insert(rpos, patternList.at(npattern));
    lock->unlock();
    emit patternChanged();
  }
  else{
    while(sequence.count()<=rpos)
      sequence.insert(sequence.count(), NULL);
    lock->unlock();
    emit sequenceSizeChanged(sequence.count());
    setPatternInSequence(rpos, npattern);
  }
}

/**
 * @short Removes one pattern from the sequence
 */
void Track::removePatternInSequence(unsigned int rpos){
  lock->lock();
  if ((rpos<sequence.count()) && (sequence.count()>1)){
    sequence.remove(rpos);
    lock->unlock();
    emit patternChanged();
    emit sequenceSizeChanged(sequence.count());
    return;
  }
  lock->unlock();
}
/**
 * @short Inserts one pattern to the the sequence
 */
void Track::insertPatternInSequence(unsigned int rpos){
  lock->lock();
  if (rpos<sequence.count()){
    sequence.insert(rpos, NULL);
    lock->unlock();
    emit patternChanged();
    emit sequenceSizeChanged(sequence.count());
    return;
  }
  lock->unlock();
}

/**
 * @short Returns the id of the current playing pattern.
 *
 * This info is needed to set correctly the step pointer on the views.
 */
int Track::currentPlayingPatternId(){
  return patternList.findRef(currentPattern);
}

/**
 * @short Gets the track Id from its position
 */
int Track::getPatternIdInSequence(int pos){
  if (((unsigned int)pos)>=sequence.count())
    return -1;
  return patternList.findRef(sequence.at(pos));
}

/**
 * @short Sets a new name fot this Track
 *
 * A signal is emmited and all aperances of the name are changed
 */
void Track::setNewName(const QString &name){
  if (name==QObject::name())
    return;
  setName(name);
  if (NULL!=midi)
    midi->setPortName(portId, name.ascii());
  emit nameChanged(this,name);
}


/**
 * @short Number of Pattern s contained in this Track
 */
int Track::npatterns(){
  return patternList.count();
}

/**
 * @short Try to set a new loop region
 */
void Track::newLoopSlot(int begin, int end){
  emit newLoop(begin,end);
}

/**
 * @short Gets port id of the ALSAMidi
 */
int Track::getPortId(){
  return portId;
}

/**
 * @short Set the alsa output to this name
 */
void Track::setALSAOutput(const QString &t){
  lock->lock();
  ALSAoutput=t;

  if (midi!=NULL){
    int cid, pid,ok;
    
    ok=sscanf(t.ascii(),"%d:%d",&cid, &pid);
    printf("Selected %d %d\n",cid,pid);
    if (ok==2)
      midi->connectTrack(this,  cid, pid);
  }
  else{
    printf("Output got (%s), but delayed\n",t.ascii());
  }
  lock->unlock();
}

/**
 * @short Returns name of the ALSA output name
 */
QString Track::getALSAOutput(){
  return ALSAoutput;
}

/**
 * @short Sets the Midi channel to channel
 */
void Track::setMIDIChannel(int channel){
  MIDIChannel=channel;
  emit setMIDIChannelSignal(channel);
}

/**
 * @short Sets the Midi program
 */
void Track::setMIDIProgram(int program){
  MIDIProgram=program;
  if (NULL!=midi)
    midi->programChange(MIDIProgram, MIDIChannel,portId);
  emit setMIDIProgramSignal(program);
}

/**
 * @short Gets the midi channel number
 */
int Track::getMIDIChannel(void){
  return MIDIChannel; 
}

/**
 * @short Sets the mute flag to on
 */
void Track::mute(){
  lock->lock();
  muteFlag=true;
  lock->unlock();
  stopNotes();
  emit muteChanged(this,muteFlag);
}

/**
 * @short Sets the mute flag to off
 */
void Track::unmute(){
  lock->lock();
  muteFlag=false;
  lock->unlock();
  emit muteChanged(this,muteFlag);
}


/**
 * @short Sends to MIDI the initial data, as the instrument to play, sysexs...
 */
void Track::sendInitialData(){
  // already locked here...
  midi->programChange(MIDIProgram, MIDIChannel,portId);
  
  SysEx *s;
  for (s=initialSysEx.first();s;s=initialSysEx.next())
    midi->sendSysEx(s->getSysEx().data(),s->length(),MIDIChannel,portId);

}

/**
 * @short Adds a initial SysEx command
 */
void Track::addInitialSysEx(SysEx *s){
  lock->lock();
  initialSysEx.append(s);
  lock->unlock();
  connect(s,SIGNAL(delSysEx(SysEx *)),
	  this, SLOT(delInitialSysEx(SysEx *)));
  emit addedInitialSysEx(s);
}

/**
 * @short Removes a initial SysEx command
 */
void Track::delInitialSysEx(SysEx *s){
  lock->lock();
  initialSysEx.remove(s);
  lock->unlock();
  emit deletedInitialSysEx(s);
}
