/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#ifndef __TRACK_H__
#define __TRACK_H__

#include <qobject.h>
#include <qptrlist.h>
#include <qvaluelist.h> 

#include "pattern.h"
#include "sysexbag.h"

class Midi;
class QMutex;

/**
 * @short Creates an empty track with the desired name.
 *
 * A track is a container of Patterns (patternList), a sequence of
 * such Patterns (sequence) and a little about playing mechanisms.
 *
 * The playing mecanisms are only the correct execution of step()
 * acording to sequence, and the correct asignation of Midi objects
 * and ports.
 */
class Track:public QObject{
  Q_OBJECT
private:
  /// Lock to keep the object thread safe (needed)
  QMutex *lock;
  /// The Midi object that this track exports to
  Midi *midi;
  /// The portId on the Midi object this tracks exports. It's not the same as MIDI channel
  int portId;
  /// Midi Channel
  int MIDIChannel;
  /// Midi instrument
  int MIDIProgram;
  /// The size of each sequence unit
  int stepSize;
  /// Size of a single beat
  int beatLength;
  /// The first step in the current Pattern, needed to play patterns from begining
  int firstStepInCurrentPattern;
  /// List of patterns
  QPtrList<Pattern> patternList;
  /**
   * @short Sequence List.  
   *
   * Each item is a new sequence that will be trigered when stepSize
   * steps pass. If the item is NULL there is no pattern at this
   * point; if previous pattern is large enought it will continue
   * sounding, if not, MIDI silence.
   */
  QPtrList<Pattern> sequence;
  /// Current playing pattern. Just a speedup
  Pattern *currentPattern;
  /// List of last played notes, transmited to the next playing pattern to stop notes corectly
  QValueList<MidiNote> lastNotesPlayed;
  /// Name of ALSA output, if any
  QString ALSAoutput;
  /// Whether is muted or not
  bool muteFlag;
  /// The SysEx Bag
  SysExBag *sysexbag;
  /// The initial SysEx commands. They are sent at the play start.
  QPtrList<SysEx> initialSysEx;
  

public:
  Track(SysExBag *syexbag, QObject *parent,const char *name);
  ~Track();
  Pattern *pattern(int n=-1);
  int npatterns();
  
  QString patternNameAtRow(int row, bool &first);

  int getStepSize();
  int getBeatLength(){ return beatLength; }
  int getSequenceSize();
  int currentPlayingPatternId();
  bool step(int step);
  int getPatternIdInSequence(int pos);
  int getPortId();
  int getMIDIChannel();
  int getMIDIProgram(){ return MIDIProgram; }

  int getFirstStepInCurrentPattern(){ return firstStepInCurrentPattern; }

  SysExBag *getSysExBag(){ return sysexbag; }
  QPtrList<SysEx> getInitialSysEx(){ return initialSysEx; }

  QString getALSAOutput();

  bool isMutted(){ return muteFlag; };

signals:
  /// A new pattern is added
  void newPattern(Pattern *newPattern);
  /// The pattern sequence configuration have changed
  void patternChanged();
  /// Try to redraw the step pointer on appropiate track...
  void redrawStepPointer(int);
  /// The sequence size changed to this
  void sequenceSizeChanged(unsigned int);
  /// The name have changed
  void nameChanged(Track *,const QString &name);
  /// Destroys current track. Indicate pointer to help things
  void deleteTrack(Track *tr);
  /// Destroys selected pattern. 
  void deletedPattern(Pattern *pt);
  /// The selecion is changed to this particular track. It must propagate the change.
  void newLoop(int begin, int end);
  /// Signal emited when a new midi channel is selected
  void setMIDIChannelSignal(int);
  /// Signal emited when a new midi program is selected
  void setMIDIProgramSignal(int);
  /// Sets the mute status
  void muteChanged(Track*,bool mute);
  /// New initial sysex added
  void addedInitialSysEx(SysEx *s);
  /// Initial sysex deleted
  void deletedInitialSysEx(SysEx *s);
  /// New signature
  void newSignature(int,int);

public slots:
  void setMidi(Midi *_midi);
  void addPattern();
  void addPattern(Pattern *newPattern);
  void delPattern(Pattern *newPattern);
  void setPatternInSequence(unsigned int rpos, unsigned int npattern);
  void insertPatternInSequence(unsigned int pos);
  void removePatternInSequence(unsigned int pos);
  void setNewName(const QString &name);
  void newLoopSlot(int begin, int end);
  void setALSAOutput(const QString &name);
  void setMIDIChannel(int channel);
  void stopNotes();
  void mute();
  void unmute();
  void sendInitialData();
  void setMIDIProgram(int prg);
  void emitDeleteTrack(Track *tr){ emit deleteTrack(tr); }

  void addInitialSysEx(SysEx *s);
  void delInitialSysEx(SysEx *s);
  void setSignature(int a,int b){ beatLength=a; stepSize=a*b; emit newSignature(a,b); };
};


#endif 
