/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include <qtimer.h>
#include <qapplication.h>
#include <stdio.h>
#include <fcntl.h>
#include <linux/rtc.h>
#include <sys/ioctl.h>
#include <sched.h>
#include <sys/resource.h>
#include <kconfig.h>
#include <kapplication.h>

#include <sys/time.h> 
#include <unistd.h> 

#include "tracker.h"
#include "track.h"

#include <iostream>
using namespace std;

/**
 * @short Creates a tracker object. 
 */
Tracker::Tracker(SysExBag *bag, QObject *parent,float BPM):
  QObject(parent, "tracker"), midi(bag,this,"kmiditrackermidi"){
  bpm=BPM;

  playing=false;
  stop=false;
  step=0;
  quit=false;
  doLoop=false;
  loopBegin=-1; loopEnd=-1;
  beats_per_bar=4;
  beat_length=4;
  lastNoteStep=-1;
  timingMethod=0;
  origTimingMethod=0;

  reloadSettings();

  KConfig *config=kapp->config();
  config->setGroup("Performance");
  if (config->readEntry("RTC","true")=="true")
    initRTC();

  midiclockcount=0;
  midiclock_dostep=false;

  start();
}

/**
 * @short Reloads the settings from the config file
 */
void Tracker::reloadSettings(){
  KConfig *config=kapp->config();
  config->setGroup("MIDI");

  MTCOffset=config->readEntry("MTCOffset","0").toInt();
  sendMTC=config->readEntry("SendMidiClock","true")=="true";
}

/** 
 * @short Opens rtc device, and sets it up 
 */
void Tracker::initRTC(){
  rtccount=0;
  printf("Opening RTC\n");
  rtcfd=open("/dev/rtc",O_RDONLY);
  if (rtcfd!=-1){
    int ok;
    rtcfreq=1024;
    ok=ioctl(rtcfd, RTC_IRQP_SET, rtcfreq);
    if (ok<0){
      perror("RTC error setting interrupt freq");
      printf(" Reverting to 64 Hz clock. \n\n"
	     "Allow users to use RTC's full power with\n"
	     "echo 1024 > /proc/sys/dev/rtc/max-user-freq\n\n");

      rtcfreq=64;
      ok=ioctl(rtcfd, RTC_IRQP_SET, rtcfreq);
      if (ok<0){
	close(rtcfd);
	rtcfd=-1;
	perror("RTC error setting interrupt freq. Reverting to old timing method");
      }
    }
    if (rtcfd!=-1){
      ok=ioctl(rtcfd, RTC_PIE_ON, 0);
      if (ok<0){
	close(rtcfd);
	rtcfd=-1;
	perror("RTC error setting on interrupt. Reverting to old timing method");
      }
    }
    timingMethod=1;
    origTimingMethod=1;
  }
  else
    perror("RTC not avaliable, using old unexact timing method");
}

/**
 * @short Graciously deletes the tracker object.
 */
Tracker::~Tracker(){
  quit=true;
  wait();
  close(rtcfd);
}

/**
 * @short Adds a track to the tracker
 */
void Tracker::addTrack(Track *t){
  RWlock.lock();
  t->setMidi(&midi);
  tracks.append(t);
  connect(this, SIGNAL(redrawStepPointer(int)), t, SIGNAL(redrawStepPointer(int)));
  connect(t, SIGNAL(deleteTrack(Track *)), this, SLOT(delTrack(Track *)));
  connect(this, SIGNAL(newSignature(int,int)), t, SLOT(setSignature(int,int)));
  t->setSignature(beatLength(), beatsPerBar());
  RWlock.unlock();
  t->newLoopSlot(loopBegin, loopEnd);
}

/**
 * @short Deletes a track from the tracker
 */
void Tracker::delTrack(Track *t){
  RWlock.lock();
  tracks.remove(t);
  RWlock.unlock();
}

/** 
 * @short Loop over steps function. It is another thread.
 *
 * Tracker thread. It waits till there is a start, and then loops over
 * and over the tracks, trigering play events when appropiate.
 */
void Tracker::run(){
  setRealTime();

  while(!quit){
    if (playing)
      doPlayStep();

    if (stop){
      RWlock.lock();
      playing=false;
      stop=false;
      step=((doLoop) && (loopBegin==-1)) ? 0 : loopBegin;
      RWlock.unlock();
      StartStopEvent *ev=new StartStopEvent(false);
      QApplication::postEvent( this, ev );
      midi.sendStopPlay();
      midi.setBlockMode(false);
      timingMethod=origTimingMethod;
    }

    switch(timingMethod){
    case 0:
      usleep(int((60.0e6/beat_length)/bpm));  /* It do not gives exact timing, to get it
				   calculate the time elapsed below and
				   substrat it to this number. It is not
				   exactly again, but more correct. FIXME */
      checkMIDIInput();
      break;
    case 1:
      waitRTC();
      break;
    case 2:
      waitMIDIClock();
      break;
    default:
      usleep(int((60.0e6/beat_length)/bpm));
      break;
    }
  }
  terminate();
}

/**
 * @short Wait for beat in the RTC way
 */
void Tracker::waitRTC(){
  int midiclockcount,newmidiclockcount;
  int rtcnewstep, rtcstep;
  double clockstepscale=((bpm*24.0)/60.0)/float(rtcfreq);
  double rtcstepscale=((bpm*beat_length)/60.0)/float(rtcfreq);
  midiclockcount=int(rtccount*clockstepscale);
  rtcstep=int(rtccount*rtcstepscale);
  do{
    // Wait for new RTC event, and add it
    int d;      
    checkMIDIInput();
    if (-1==read(rtcfd, &d, sizeof(d))){
      perror("Can't read correctly RTC. Reverting to old timer");
      close(rtcfd);	  rtcfd=-1;
      timingMethod=0;
      return;
    }
    rtccount++;
    
    // Check if we change to another midi clock message
    newmidiclockcount=int(rtccount*clockstepscale);
    if (midiclockcount!=newmidiclockcount){
      midiclockcount=newmidiclockcount;
      if (playing && sendMTC)
	midi.sendMIDIClock();
    }
    
    // And if we change the beat
    rtcnewstep=int(rtccount*rtcstepscale);
  }while(rtcnewstep==rtcstep);
}

/**
 * @short Wait for beat in the MIDI Clock way
 */
void Tracker::waitMIDIClock(){
  midiclock_dostep=false;
  do{
    midi.updateMIDIInputBlock(this);
  }while((!midiclock_dostep) && (timingMethod==2));

  // Guess what bpm we are receiving
  if (midiTempoGuessBeat>=beat_length){
    midiTempoGuessBeat=0;
    struct timeval end;
    gettimeofday(&end,NULL);
    double t,t2;
    t=(end.tv_sec*1e6+end.tv_usec);
    t2=(t-midiTempoGuessTime)*1e-6;
    midiTempoGuessTime=t;

    // in t we calculate tempo
    t=60.0/t2;

    QEvent *ev=new TempoEvent(t);
    QApplication::postEvent( this, ev );

  }
  midiTempoGuessBeat++;
}

/**
 * @short A MIDI Clock was received
 */
void Tracker::midiClock(){
  if (timingMethod!=2){
    //printf("New timing method MIDI Clock!\n");
    origTimingMethod=timingMethod;
    timingMethod=2;

    struct timeval start;
    gettimeofday(&start,NULL);
    midiTempoGuessTime=(start.tv_sec*1e6+start.tv_usec);
    midiTempoGuessBeat=1;
  }
  midiclockcount++;
  if (playing && sendMTC)
    midi.sendMIDIClock();
  // Only works with 1, 2, 3, 4, 6, 8, 12, and 24 for beat_length
  if (((midiclockcount)%(24/beat_length))==0)
    midiclock_dostep=true;
}


/**
 * @short Reverts timing method to the orig one
 */
void Tracker::revertTimingMethod(){
  puts("Reverting to previous method\n");
  timingMethod=origTimingMethod;
}

/**
 * @short Plays a simple step of the song
 *
 * It calls all the needed track->step() to make them produce midi
 * events.
 */
void Tracker::doPlayStep(){
  RWlock.lock();
  /* set step counter in a correct state */
  if (doLoop && loopBegin!=-1 && loopEnd!=-1)
    step=((step-loopBegin+(loopEnd+1-loopBegin))%(loopEnd+1-loopBegin))+loopBegin;

  int sequenceStep=(step/(beatsPerBar()*beatLength()));
  /* step on each track */
  Track *i;
  for(i=tracks.first();i;i=tracks.next()){
    i->step(step);
    // FIXME. Maybe it takes too many cycles on a 
    // protected and rt area... i guess it dont.
    if (i->getSequenceSize()<=sequenceStep && step!=0){
      if (doLoop){
        RWlock.unlock();
        step=0;
        doPlayStep();
        return;
      }
      else{
        RWlock.unlock();
        stopPlay();
        return;
      }
    }
  }

  /* increment pointer and set to correct state */
  step++;
  if (doLoop && loopBegin!=-1 && loopEnd!=-1)
    step=((step-loopBegin+(loopEnd+1-loopBegin))%(loopEnd+1-loopBegin))+loopBegin;
  StepEvent *ev=new StepEvent(step);
  RWlock.unlock();

  QApplication::postEvent( this, ev );
}


/**
 * @short Checks the midi inputs for incomming data
 *
 * If an event is received, it is resent to the trackerbox (parent) as
 * an event.
 */
void Tracker::checkMIDIInput(){
  midi.updateMIDIInput(this);
}
/**
 * @short Adds a MIDI note at current position
 */

void Tracker::addNote(const MidiNote &note){
  MidiNoteEvent *ev=new MidiNoteEvent(note);
  QApplication::postEvent( this, ev );
}

/**
 * @short Start playing
 */
void Tracker::startPlay(){
  RWlock.lock();
  if (!playing){
    playing=true;

    Track *track;
    for (track=tracks.first();track;track=tracks.next())
      track->sendInitialData();

    StartStopEvent *ev=new StartStopEvent(true);
    QApplication::postEvent( this, ev );
    midi.sendStartPlay();

    rtccount=MTCOffset*((bpm*beat_length)/60.0)/float(rtcfreq);
    midiclockcount=MTCOffset;
    step=0;
  }
  RWlock.unlock();
}

/**
 * @short Stops the tracker
 */
void Tracker::stopPlay(){ 
  RWlock.lock();
  if (playing){
    stop=true;
  }
  timingMethod=origTimingMethod;
  step=-1;
  RWlock.unlock();
}

/**
 * @short Gets the step number.
 */
int Tracker::getStep(void){
  int r;
  RWlock.lock();
  r=step;
  RWlock.unlock();
  return r;
}

/**
 * @short Set the step number
 */
void Tracker::setStep(int newStep){
  RWlock.lock();
  step=newStep;
  RWlock.unlock();
}

/**
 * @short Sets the BPM
 */
void Tracker::setBPM(double BPM){
  RWlock.lock();
  bpm=BPM;
  RWlock.unlock();
  emit newBPM(BPM);
}

/**
 * @short Activates the loop
 */
void Tracker::activateLoop(bool loop){
  RWlock.lock();
  //  printf("Loop %d\n",loop);
  doLoop=loop;
  if (loopBegin!=-1 && loopEnd!=-1)
    step=((step-loopBegin+(loopEnd+1-loopBegin))%(loopEnd+1-loopBegin))+loopBegin;
  RWlock.unlock();
  emit loopActivated(loop);
}

/**
 * @short Sets the new loop interval
 *
 * The new interval will be used if the doLoop bool is true.
 */
void Tracker::newLoop(int begin, int end){
  RWlock.lock();
  //  printf("New loop selected %d %d\n",begin,end);
  loopBegin=begin;
  loopEnd=end;
  if (doLoop)
    step=loopBegin;
  RWlock.unlock();
  emit loopChanged(begin,end);
}

/**
 * @short Sets the loop region and activated option
 *
 * Calls real changers, and emit signal
 */
void Tracker::setLoopRegion(int b,int e,bool a){
  newLoop(b,e);
  activateLoop(a);
  emit newLoopRegion(b,e,a);
}


/**
 * @short Manages StepEvent
 *
 * Each time a step is done an event is trigered, and this function
 * captures it and transform it to a signal. This way all the
 * mechanism is thread safe.
 */
void Tracker::customEvent (QCustomEvent *ev){
  switch(ev->type()){
  case QEvent::User+1:
    {
      int step;
      if (isPlaying()){
	step=getStep()-1;
	if (-1==step && isLoopActive())
	  step=getLoopEnd();
      }
      else
	step=-1;
      // If not deleting note just after placing one
      if (!((((MidiNoteEvent*)ev)->note()==MidiNote::StopNote) && (lastNoteStep==step)))
	emit midiNoteSignal(step,*((MidiNoteEvent*)ev));
      if (isPlaying())
	lastNoteStep=step;
    }
    break;
  case QEvent::User+2: 
    emit redrawStepPointer(((StepEvent*)ev)->getStep());
    break;
  case QEvent::User+3: 
    if (((StartStopEvent*)ev)->isStart())
      emit startSignal();
    else
      emit stopSignal();
    break;
  case QEvent::User+4: 
    setBPM(((TempoEvent*)ev)->getTempo());
    break;
  default:
    break;
  }
}

/**
 * @short Sets realtime scheduler
 */
void Tracker::setRealTime(){
  KConfig *config=kapp->config();
  config->setGroup("Performance");
  struct sched_param schp;
  memset(&schp, 0, sizeof(schp));
  schp.sched_priority = sched_get_priority_max(SCHED_FIFO);

  if (config->readEntry("RealTime","true")=="true")
    if (sched_setscheduler(0, SCHED_FIFO, &schp) != 0) {
      perror("\nsched_setscheduler");
      fprintf(stderr," consider doing something to make me realtime (like realtime-lsm)\n\n");
    }
    else
      printf("Using Realtime Scheduling\n");

  /* Set higher priority, if posible */
  if (config->readEntry("HighPriority","true")=="true")
    if (setpriority(PRIO_PROCESS, 0, -5)!=0)
      perror("setpriority");
    else
      printf("Using high priority\n");
}


