/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#ifndef __TRACKER_H__
#define __TRACKER_H__

#include <qthread.h>
#include <qptrlist.h> 
#include <qobject.h>
#include <qevent.h>
#include "alsamidi.h"

class Track;
class QTimer;
class SysExBag;

/**
 * @short Step event
 *
 * A step geos, and an event is triggered.
 */
class StepEvent:public QCustomEvent{
 public:
  StepEvent(int _step):QCustomEvent(QEvent::User+2){step=_step;};
  int getStep(){ return step;};
 protected:
  int step;
};


/**
 * @short Start or Stop event
 *
 * A start or stop happened
 */
class StartStopEvent:public QCustomEvent{
 public:
  StartStopEvent(bool _isStart):QCustomEvent(QEvent::User+3){mIsStart=_isStart;};
  int isStart(){ return mIsStart;};
  int isStop(){ return !mIsStart;};
 protected:
  bool mIsStart;
};


/**
 * @short New tempo event
 *
 * Tempo have changed to this...
 */
class TempoEvent:public QCustomEvent{
 public:
  TempoEvent(float newTempo):QCustomEvent(QEvent::User+4){tempo=newTempo;};
  float getTempo(){ return tempo; };
 protected:
  float tempo;
};


/**
 * @short The real tracker. Spawns a thread
 *
 * It is the real tracker that spawns a thread to send correct steps
 * to each track. Here you can change BPM, play, pause or stop, thread
 * safe.
 */
class Tracker:public QObject,QThread{
  Q_OBJECT
public:
  /// Creates the new tracker thread 
  Tracker(SysExBag *bag, QObject *parent,float bpm=90);
  /// Safely destroys unused data.
  ~Tracker();
  /// Call fo the new thread.
  virtual void run(void);
  /// Gets step number.
  int getStep();
  /// Returns BPM
  float getBPM(){return bpm;}
  int getLoopBegin(){return loopBegin;}
  int getLoopEnd(){return loopEnd;}
  bool isLoopActive(){ return doLoop; }
  Midi *getMidi(){ return &midi; }
  bool isPlaying(){ return playing; }
  int beatLength(){ return beat_length; }
  int beatsPerBar(){ return beats_per_bar; }

  void midiClock();
  void revertTimingMethod();

protected:
  void customEvent (QCustomEvent *);
  void checkMIDIInput();
  void doPlayStep();

  void initRTC();
  void waitRTC();

  void waitMIDIClock();

  void setRealTime();

  /// Current bpm
  float bpm;
  /// Beats per bar
  int beats_per_bar;
  /// Beat length
  int beat_length;
  /// Step at witch last note was added. To avoid NoteOff when already
  /// a noteon is in place
  int lastNoteStep;

  /// MTC Offset
  int MTCOffset;
  /// Send MTC
  bool sendMTC;


  /// Exclusive access to internal variables. USed for thread safety.
  QMutex RWlock;
  /// Whether it is laying or not.
  bool playing;
  /// Whether it should stop on next step
  bool stop;
  /// Number of actual step. Tracks get this number and module it.
  int step;
  /// A list of tracks to call when a new step should be launched.
  QPtrList<Track> tracks;
  /// The MIDI interface
  Midi midi;
  /// Indicates that the play thread must end asap
  bool quit;
  /// Whetehr do loop or not
  bool doLoop;
  /// Loop region 
  int loopBegin, loopEnd;
  /// Timing Method 0 usleep, 1 RTC, 2 MIDI Clock
  int timingMethod;
  /// @{ @name RTC stuff
  /// rtc descriptor
  int rtcfd;
  /// rtc frecuency, higher more precission
  int rtcfreq;
  /// RTC beats counter
  int rtccount;
  /// @}
  /// @{ @name MIDI Clock method
  /// Whether a new step happened
  bool midiclock_dostep;
  /// MIDI clock count. Also used in RTC Timing mode to send correct
  /// midi clock signals
  int midiclockcount;
  /// Original timing method. When stop the timing is reverted
  int origTimingMethod;
  /// To guess tempo, start time taking notes
  double midiTempoGuessTime;
  /// Number of beats guessing
  int midiTempoGuessBeat;
  /// @}

signals:
  /// Redraw the mark of the step pointer
  void redrawStepPointer(int step);
  /// Signal emited when stop.
  void startSignal();
  /// Signal emited when stop.
  void stopSignal();
  /// The new selected loop region
  void loopChanged(int,int);
  /// Emits the new BPM
  void newBPM(double);
  /// A new note is pressed on the midi input
  void midiNoteSignal(int step, const MidiNote &);
  /// The loop have been activated
  void loopActivated(bool loop);
  /// The loop region should be changed to [begin,end] and activate
  void newLoopRegion(int begin,int end, bool activate);
  /// The signature changed to this
  void newSignature(int a,int b);

public slots:
  /// Adds a new track
  void addTrack(Track *);
  /// Deletes a new track
  void delTrack(Track *);
  /// Strats playing
  void startPlay();
  /// Stops playing
  void stopPlay();
  /// Sets actual step. To jump position.
  void setStep(int newStep);
  /// Sets a new BPM
  void setBPM(double BPM);
  /// Sets the region and activation
  void setLoopRegion(int,int,bool);
  /// The new region loop
  void newLoop(int,int);
  /// Activate the loop
  void activateLoop(bool);
  /// Sets current signature
  void setSignature(int a,int b){ beat_length=a; beats_per_bar=b; emit newSignature(a,b); }
  /// Adds a new note
  void addNote(const MidiNote &);
  /// Reload settings
  void reloadSettings();
};

#endif
