/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/

#include "trackerbox.h"
#include "track.h"
#include "tracker.h"
#include "signaturewidget.h"

#include <qlabel.h>
#include <qbutton.h>
#include <kmainwindow.h>
#include <qapplication.h>
#include <knuminput.h> 
#include <kaction.h>

#include <klocale.h>

#include <iostream>
using namespace std;



/**
 * @short Creates the TrackerBox from a KMainWindow
 *
 * At first the TrackerBox creates several KActions from widgets that
 * depend of the KMainWindow but are not shown. KActions can be
 * changed as score plays, and some other events.
 */
TrackerBox::TrackerBox(SysExBag *bag, KMainWindow *parent, 
		       const char *name):QWidget(parent,name){
  tracker=new Tracker(bag, this);

  playstop=new KAction(i18n("&PlayStop"),"player_play",0,this, SLOT(playStop()),
		       parent->actionCollection(),"playstop_song");
  play=new KAction(i18n("&Play"),"player_play",0,this, SIGNAL(startPlay()),
		   parent->actionCollection(),"play_song");
  stop=new KAction(i18n("&Stop"),"player_stop",0,this, SIGNAL(stopPlay()),
		   parent->actionCollection(),"stop_song");
  playLoop=new KToggleAction(i18n("Activate &Loop"),"reload",0,
			     parent->actionCollection(),"activate_loop");
  connect(playLoop, SIGNAL(toggled(bool)), this, SIGNAL(activateLoop(bool)));

  signatures=new SignatureWidget(tracker->beatLength(),
				 tracker->beatsPerBar(),
				 this,"signature_combo");
  connect(signatures,SIGNAL(newSignature(int,int)),
	  tracker,SLOT(setSignature(int,int)));
  connect(tracker,SIGNAL(newSignature(int,int)),
	  signatures,SLOT(updateSignature(int,int)));
  signatures->hide();
  new KWidgetAction(signatures, "Signature", 0,0,0, 
		    parent->actionCollection(),"signature_combo");

    
  sp=new KDoubleSpinBox(1.0, 1000.0, 1.0, tracker->getBPM(), 2, this);
  sp->hide();
  new KWidgetAction(sp,"BPM Selector", 0, 0, 0, parent->actionCollection(), "bpm_selector");

  time=new QLabel("<font color=\"blue\">00:00:00</font>",this);
  time->hide();
  new KWidgetAction(time,"Time Label", 0, 0, 0, parent->actionCollection(), "time_label");
  
  connect(sp,SIGNAL(valueChanged(double)),tracker,SLOT(setBPM(double)));
  connect(this, SIGNAL(startPlay()), tracker, SLOT(startPlay()));
  connect(this, SIGNAL(stopPlay()), tracker, SLOT(stopPlay()));
  connect(this, SIGNAL(activateLoop(bool)), tracker, SLOT(activateLoop(bool)));

  connect(tracker, SIGNAL(loopActivated(bool)), this, SLOT(loopActivated(bool)));
  connect(tracker, SIGNAL(redrawStepPointer(int)), this, SLOT(stepSlot(int)));
  connect(tracker, SIGNAL(newBPM(double)), this, SLOT(newBPM(double)));

  connect(tracker, SIGNAL(startSignal()), this, SLOT(setPlaying()));
  connect(tracker, SIGNAL(stopSignal()), this, SLOT(setNotPlaying()));
}


/**
 * @short Deleted associated Tracker
 */
TrackerBox::~TrackerBox(){
  delete sp;
  delete time;
  delete tracker;
}

/**
 * @short Adds a new track to the TrackList.
 * 
 * This action is delegated to the Tracker
 */
void TrackerBox::addTrack(Track *track){
  tracker->addTrack(track);
}

/**
 * @short Deletes a track from the TrackList.
 * 
 * This action is delegated to the Tracker
 */
void TrackerBox::delTrack(Track *track){
  tracker->delTrack(track);
}

/**
 * @short Gets actual Beats Per Minute
 */
float TrackerBox::getBPM(){
  return tracker->getBPM();
}

/**
 * @short Gets th position where loop begins
 */
int TrackerBox::getLoopBegin(){
  return tracker->getLoopBegin();
}

/**
 * @short Gets end position of the loop
 */
int TrackerBox::getLoopEnd(){
  return tracker->getLoopEnd();
}

/**
 * @short Changes BPM to bpm
 */
void TrackerBox::newBPM(double bpm){
  sp->setValue(bpm);
}

/**
 * @short Whether or not the Loop is active
 */
bool TrackerBox::isLoopActive(){
  return tracker->isLoopActive();
}

/**
 * @short A new step have happened, and some data should change
 * 
 * Time label is changed to reflect new time
 */
void TrackerBox::stepSlot(int step){
  char t[256];
  int a=(tracker->beatLength()*tracker->beatsPerBar());
  int b=tracker->beatLength();
  int c=tracker->beatsPerBar();
  step--;
  if (step>=0)
    sprintf(t,"<font color=\"darkgreen\">%02d:%02d:%02d</font>",
	    1+step/a, 
	    1+(step%a)/b, 1+step%b);
  else
    sprintf(t,"<font color=\"blue\">00:00:00</font>");
  time->setText(t);
}

/**
 * @short Gets related Midi object from Tracker
 */
Midi *TrackerBox::getMidi(){ 
  return tracker->getMidi(); 
}

/**
 * @short Sets the status to playing
 */
void TrackerBox::setPlaying(){
  play->setEnabled(false);
  stop->setEnabled(true);
}

/**
 * @short Used to set the status to not playing
 */
void TrackerBox::setNotPlaying(){
  play->setEnabled(true);
  stop->setEnabled(false);
  stepSlot(0);
}

/**
 * @short Sets the loop region and if it's on of off
 */
void TrackerBox::loopActivated(bool loop){
  playLoop->setChecked(loop);
}

/**
 * @short Gets the associated Tracker
 */
Tracker *TrackerBox::getTracker(){
  return tracker;
}

/**
 * @short Play/Stop (toggle)
 */
void TrackerBox::playStop(){
  if (tracker->isPlaying())
    emit stopPlay();
  else
    emit startPlay();
}
