/***************************************************************************
 *   Copyright (C) 2004 by David Moreno Montero                            *
 *   david@monasteriomono.org                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA 02110-1301, USA.           *
 ***************************************************************************/


#include <qsplitter.h>
#include <qwidgetstack.h>
#include <qlineedit.h>
#include <qspinbox.h>
#include <qtooltip.h>
#include <qcombobox.h>
#include <qptrlist.h>
#include <qpushbutton.h>
#include <qlayout.h>
#include <qtoolbox.h>
#include <qgroupbox.h>
#include <qcursor.h>
#include <qapplication.h>

#include <klocale.h>
#include <klistbox.h>
#include <kmessagebox.h>
#include <kpopupmenu.h>
#include <kconfig.h>
#include <kapplication.h>

#include "sysex.h"
#include "trackview.h"
#include "patternview.h"
#include "sequenceview.h"
#include "alsamidi.h"

/**
 * @short Creates a new trackview and the track inside
 *
 * Creates the views of the several components of a TrackView and
 * connects signals.
 */
TrackView::TrackView(Track *tr, QWidget *parent, const char *nname):QVBox(parent,nname){
  track=tr;
  constructView();
  
  reloadSettings();
}


/**
 * @short Creates the real view of the track
 *
 * It is created using the data from the Track track.
 */
void TrackView::constructView(){
  int i;
  QSplitter *vbox;
  QHBox *hbox;
  QPushButton *button;


  view=new QSplitter(Qt::Horizontal,this,"track_view_2");
  patternView=new QWidgetStack(view, "pattern_view_stack");
  connect(patternView, SIGNAL(aboutToShow(QWidget*)), 
	  this, SLOT(aboutToShowPatternView(QWidget*)));
  currentPlayingView=NULL;


  vbox=new QSplitter(Qt::Vertical,view,"track_view_2");
  tbox=new QToolBox(vbox);

  /*
   * Track Properties
   */
  QWidget *properties=new QWidget(tbox);
  QVBoxLayout *layout=new QVBoxLayout(properties,1,1);
  //  properties->setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Fixed);

  /// Some properties...
  QWidget *t;
  QGroupBox *gbox;
  gbox=new QGroupBox(i18n("Track name:"),properties);
  gbox->setColumns(1);
  t=new QLineEdit(name(), gbox, "track_name_edit");
  connect(t,SIGNAL(textChanged(const QString&)), track, SLOT(setNewName(const QString&)));
  QToolTip::add(t,i18n("Select the name of the track. This is also the name exported to ALSA"));
  layout->addWidget(gbox);

  gbox=new QGroupBox(i18n("Base Octave:"), properties);
  gbox->setColumns(1);
  baseOctave=new QSpinBox(1,5,1, gbox, "base_octave");
  baseOctave->setValue(3);
  layout->addWidget(gbox);

  hbox=new QHBox(properties);
  gbox=new QGroupBox(i18n("Midi Channel:"), hbox);
  gbox->setColumns(1);
  QSpinBox *midiChannel=new QSpinBox(1,16,1, gbox, "midi_channel");
  midiChannel->setValue(track->getMIDIChannel());
  connect(midiChannel,SIGNAL(valueChanged(int)),track,SLOT(setMIDIChannel(int)));
  connect(track, SIGNAL(setMIDIChannelSignal(int)), midiChannel, SLOT(setValue(int)));
  
  gbox=new QGroupBox(i18n("Midi Program:"), hbox);
  gbox->setColumns(1);
  QSpinBox *midiProgram=new QSpinBox(1,128,1, gbox, "midi_program");
  midiProgram->setValue(track->getMIDIProgram());
  connect(midiProgram,SIGNAL(valueChanged(int)),track,SLOT(setMIDIProgram(int)));
  connect(track, SIGNAL(setMIDIProgramSignal(int)), midiProgram, SLOT(setValue(int)));

  layout->addWidget(hbox);
  
  gbox=new QGroupBox(i18n("MIDI output: "), properties);
  gbox->setColumns(1);
  alsaCombo=new QComboBox(gbox, "track_combo"); 
  alsaCombo->insertItem(i18n("None"));
  QToolTip::add(alsaCombo,i18n("Select to which ALSA MIDI input should this track output.\n"
			       "None lets the connection to an external program."));
  connect(alsaCombo,SIGNAL(activated(const QString &)),
	  this,SLOT(alsaComboSelected(const QString &)));
  //alsaCombo->setMaximumWidth(200);
  layout->addWidget(gbox);
 

  gbox=new QGroupBox(i18n("Initial SysEx"),properties);
  gbox->setColumns(2);
  gbox->setOrientation(Qt::Vertical);
  initialSysEx=new QListBox(gbox,"sysex_initial_list");
  QPtrList<SysEx> initial=track->getInitialSysEx();
  SysEx *ini;
  for (ini=initial.first();ini;ini=initial.next())
    initialSysEx->insertItem(ini->getName());
  hbox=new QHBox(gbox);
  hbox->setSpacing(6);
  button=new QPushButton(i18n("Add SysEx..."),hbox);
  connect(button,SIGNAL(clicked()),
	  this,SLOT(addInitialSysEx()));
  button=new QPushButton(i18n("Remove SysEx"),hbox);
  connect(button,SIGNAL(clicked()),
	  this,SLOT(delSelectedInitialSysEx()));
  layout->addWidget(gbox);

  layout->addItem(new QSpacerItem
		  (20, 1, QSizePolicy::Minimum, QSizePolicy::Expanding ));

  tbox->addItem(properties,i18n("Track Properties"));


  /*
   * Pattern properties
   */
  properties=new QWidget(tbox);
  layout=new QVBoxLayout(properties,6,6);
  gbox=new QGroupBox(i18n("Pattern name: "), properties);
  gbox->setColumns(1);
  patternNameEdit=new QLineEdit(name(), gbox, "pattern_name_edit");
  QToolTip::add(patternNameEdit,i18n("Select the name of the pattern."));
  connect(patternNameEdit,SIGNAL(textChanged(const QString&)), 
	  this, SLOT(changeCurrentPatternName(const QString&)));
  layout->addWidget(gbox);


  hbox=new QHBox(properties);
  gbox=new QGroupBox(i18n("Rows:"), hbox);
  gbox->setColumns(1);
  rowsSpin=new QSpinBox(1,2048,1, gbox, "rows_spin");
  if (currentPlayingView!=NULL)
    rowsSpin->setValue(currentPlayingView->getPattern()->getRows());
  gbox=new QGroupBox(i18n("Columns:"), hbox);
  gbox->setColumns(1);
  columnsSpin=new QSpinBox(1,128,1, gbox, "columns_spin");
  if (currentPlayingView!=NULL)
    columnsSpin->setValue(currentPlayingView->getPattern()->getCols());

  layout->addWidget(hbox);

  button=new QPushButton(i18n("Change Rows and Columns"),properties);
  connect(button, SIGNAL(clicked()),
	  this, SLOT(changeRowsAndColumns()));
  layout->addWidget(button);


  button=new QPushButton(i18n("Controllers Properties"),properties);
  connect(button, SIGNAL(clicked()),
	  this, SLOT(showControllersProperties()));
  layout->addWidget(button);


  layout->addItem(new QSpacerItem
		  (20, 1, QSizePolicy::Minimum, QSizePolicy::Expanding ));

  tbox->addItem(properties,i18n("Pattern Properties"));



  /*
   * Pattern List
   */
  properties=new QWidget(tbox);
  layout=new QVBoxLayout(properties,6,6);

  patternList=new KListBox(properties,"pattern_list");
  layout->addWidget(patternList);

  hbox=new QHBox(properties);
  hbox->setSpacing(6);
  button=new QPushButton(i18n("&New"), hbox);
  connect(button, SIGNAL(clicked()),
	  this, SLOT(addPattern()));
  button=new QPushButton(i18n("&Copy"),hbox);
  connect(button, SIGNAL(clicked()),
	  this, SLOT(copyPattern()));
  button=new QPushButton(i18n("&Remove..."), hbox);
  connect(button, SIGNAL(clicked()),
	  this, SLOT(delSelectedPattern()));
  layout->addWidget(hbox);


  tbox->addItem(properties,i18n("Pattern List"));
  
  /*
   * Sequence view
   */
  gbox=new QGroupBox(i18n("Sequence View"), vbox);
  gbox->setColumns(1);

  sequenceView=new SequenceView(false,gbox, "sequence_view");
  //tbox->addItem(sequenceView,i18n("Sequence"));


  //  tbox->setCurrentIndex(3);

  /// correct size... 
  QValueList<int> viewSize;
  viewSize.append(700);
  viewSize.append(400);
  view->setSizes(viewSize);
  
  
  /// Add patterns...
  for (i=0;i<track->npatterns();i++)
    newPatternAdded(track->pattern(i));
  sequenceView->manageTrack(track);

  /// Connect all signals
  connect(track, SIGNAL(newPattern(Pattern *)), this, SLOT(newPatternAdded(Pattern *)));
  connect(patternList, SIGNAL(highlighted(int)), patternView, SLOT(raiseWidget(int)));
  connect(sequenceView, SIGNAL(setActivePatternViewInSequence(int)), 
	  this, SLOT(selectedPattern(int)));
  connect(sequenceView, SIGNAL(positionChanged(unsigned char, unsigned char)),
	  this, SIGNAL(positionChanged(unsigned char, unsigned char)));

  connect(this, SIGNAL(patternChanged()), track, SIGNAL(patternChanged()));
  connect(track,SIGNAL(redrawStepPointer(int)), this, SLOT(redrawStepPointer(int)));

  connect(sequenceView, SIGNAL(newLoop(int,int)), this, SIGNAL(newLoop(int,int)));
  connect(track, SIGNAL(newLoop(int,int)), this, SIGNAL(newLoop(int,int)));
  connect(this, SIGNAL(loopChanged(int,int)), sequenceView, SLOT(loopChanged(int,int)));
  connect(track, SIGNAL(newSignature(int,int)), sequenceView, SLOT(setSignature(int,int)));

  connect(track, SIGNAL(deleteTrack(Track *)), this, SLOT(trackDeleted(Track *)));

  connect(this, SIGNAL(setALSAOutput(const QString &)), 
	  track, SLOT(setALSAOutput(const QString&)));

  connect(this, SIGNAL(addInitialSysEx(SysEx *)),
	  track, SLOT(addInitialSysEx(SysEx *)));
  connect(track, SIGNAL(addedInitialSysEx(SysEx *)),
	  this, SLOT(addedInitialSysEx(SysEx *)));
  connect(this, SIGNAL(delInitialSysEx(SysEx *)),
	  track, SLOT(delInitialSysEx(SysEx *)));
  connect(track, SIGNAL(deletedInitialSysEx(SysEx *)),
	  this, SLOT(deletedInitialSysEx(SysEx *)));

  connect(track, SIGNAL(deletedPattern(Pattern *)),
	  this, SLOT(deletedPattern(Pattern *)));
}


/**
 * @short Destroys no more used data
 */
TrackView::~TrackView(){

}

/**
 * @short Sets the name of a specified pattern by id.
 */
void TrackView::setPatternName(int id, const QString &name){
  disconnect(patternNameEdit,SIGNAL(textChanged(const QString&)), 
	     this, SLOT(changeCurrentPatternName(const QString&)));
  patternNameEdit->setText(name);
  connect(patternNameEdit,SIGNAL(textChanged(const QString&)), 
	  this, SLOT(changeCurrentPatternName(const QString&)));

  patternList->changeItem(QString("%1. %2").arg(id).arg(name),id);
  emit patternChanged();
}

/**
 * @short Adds a new pattern to the track
 */
void TrackView::addPattern(void){
  track->addPattern();
}

/**
 * @short Copys actual pattern to a new one
 *
 */
void TrackView::copyPattern(void){
  PatternView *pt=(PatternView*)patternView->visibleWidget();
  if (NULL!=pt){
    Pattern *p=pt->getPattern();
    Pattern *np=new Pattern(*p);
    np->setName(QString(p->name())+" copy");
    track->addPattern(np);
  }
}


/**
 * @short Slot to add a new pattern to this view.
 *
 * Just creates a new PatternView and adds it to the widget stack, the
 * entry on the pattern list..
 */
void TrackView::newPatternAdded(Pattern *pattern){
  PatternView *newPatternView;

  newPatternView=new PatternView(pattern, this, view, "pattern_view");
  newPatternView->setPatternId(patternList->count());
  patternView->addWidget(newPatternView, patternList->count());
  aboutToShowPatternView(newPatternView);
  
  char letterId=patternList->count();
  if (letterId<10)
    letterId='0'+letterId;
  else
    letterId='a'+letterId-10;

  patternList->insertItem(QString("%1. %2").
			  arg(letterId).arg(pattern->name()), 
			  patternList->count());
  patternList->setCurrentItem(patternList->count()-1);

  patternView->raiseWidget(newPatternView);
  patternNameEdit->setText(pattern->name());

  connect(baseOctave, SIGNAL(valueChanged(int)), newPatternView, SLOT(setBaseOctave(int)));

  connect(newPatternView, SIGNAL(positionChanged(unsigned char, unsigned char)),
	  this, SIGNAL(positionChanged(unsigned char, unsigned char)));

  connect(this, SIGNAL(reloadSettingsSignal()),
	  newPatternView, SLOT(reloadSettings()));
  connect(this, SIGNAL(reloadSettingsSignal()),
          this, SLOT(reloadSettings()));

  statusMessage(QString("New pattern added: %1").arg(pattern->name()),2000);
}

/**
 * @short Reloads some settings that may affect this view
 */
void TrackView::reloadSettings(){
  KConfig *config=kapp->config();
  config->setGroup("General");

  followSong=config->readEntry("FollowSong","false")=="true";
}


/**
 * @short Set focus to the selected PatternView.
 *
 * It can be any kind of QWidget too.
 */
void TrackView::aboutToShowPatternView(QWidget *w){
  PatternView *pt=(PatternView*)w;
  if (NULL!=pt){
    disconnect(patternNameEdit,SIGNAL(textChanged(const QString&)), 
	       this, SLOT(changeCurrentPatternName(const QString&)));
    patternNameEdit->setText(pt->getPatternName());
    rowsSpin->setValue(pt->getPattern()->getRows());
    baseOctave->setValue(pt->getBaseOctave());
    columnsSpin->setValue(pt->getPattern()->getCols());
    connect(patternNameEdit,SIGNAL(textChanged(const QString&)), 
	    this, SLOT(changeCurrentPatternName(const QString&)));
  }
}

/**
 * @short Redraws the pointer of the current position
 *
 * It just sets a mark on current step
 */
void TrackView::redrawStepPointer(int ){
  if (NULL!=currentPlayingView)
    currentPlayingView->clearStepPosition();

  currentPlayingView=(PatternView*)patternView->widget(track->currentPlayingPatternId());
  
  if (followSong)
    patternView->raiseWidget(currentPlayingView);
  if (NULL!=currentPlayingView)
    currentPlayingView->drawStepPosition();
}

/**
 * @short The track is deleted, so should be the view.
 */
void TrackView::trackDeleted(Track *tr){
  if (tr==track) // Just a check...
    delete this;
}

/**
 * @short Sets the selected pattern in the pattern list 
 *
 * I can't use directly a signal/slot because inexplicably the
 * desired slot does not exist.
 */
void TrackView::selectedPattern(int id){
  patternList->setSelected(id,true);
  statusMessage(QString("Selected pattern: %1").arg(id),500);
}

/**
 * @short Sets the focus to the PatternView
 */
void TrackView::setActivePatternView(){
  patternView->visibleWidget()->setFocus();
  statusMessage(QString("Selected pattern editor"),500);
}

/**
 * @short Sets the focus to the Pattern list
 */
void TrackView::setActivePatternList(){
  tbox->setCurrentItem(tbox->item(2));
  patternList->setFocus();
  statusMessage(QString("Selected pattern list"),500);
}

/**
 * @short Sets the focus to the SequenceView
 */
void TrackView::setActiveSequenceView(){
  sequenceView->setFocus();
  statusMessage(QString("Selected sequence view"),500);
}

/**
 * @short A MidiNote event that should be added to current Pattern
 *
 * Usually signaled from ALSAMidi->KMidiTrackerView
 */
void TrackView::midiNote(int step,const MidiNote &ev){
  PatternView *pt=(PatternView*)patternView->visibleWidget();
  if (NULL!=pt){
    pt->addMidiNote(step-track->getFirstStepInCurrentPattern(),ev);
  }
}

/**
 * @short Changes current pattern name to nname
 */
void TrackView::changeCurrentPatternName(const QString &nname){
  PatternView *pt=(PatternView*)patternView->visibleWidget();
  if (NULL!=pt)
    pt->setName(nname);
  statusMessage(QString("New pattern name: %1").arg(nname),500);
}

/**
 * @short Sets the current alsaportlist (the combo) to l
 *
 * This slot is signaled only at creation, and when ALSA ports changes
 * (names, additions and supressions).
 */
void TrackView::ALSAPortList(QPtrList<ALSAPort> &l){
  int cid=-1, pid=-1,ok;

  ok=sscanf(track->getALSAOutput().ascii(),"%d:%d",&cid, &pid);

  alsaCombo->clear();
  alsaCombo->insertItem(i18n("None"));

  ALSAPort *i;
  for(i=l.first();i!=NULL;i=l.next()){
    QString t=QString("%1:%2 %3:%4").arg(i->client).arg(i->port).
      arg(i->client_name).arg(i->port_name);
    alsaCombo->insertItem(t);
    if (i->client==cid && i->port==pid)
      alsaCombo->setCurrentText(t);
  }
  statusMessage(QString("Alsa ports changed"),1500);
}

/**
 * @short A ALSA client:port is selected, and the connection should be
 * made
 */
void TrackView::alsaComboSelected(const QString &t){
  int cid, pid,ok;

  ok=sscanf(t.ascii(),"%d:%d",&cid, &pid);
  printf("Selected %d %d\n",cid,pid);
  if (ok!=2)
    cid=pid=-1;
  emit setALSAOutput(track,cid,pid);
  emit setALSAOutput(t);
  statusMessage(QString("Selected %1 as output for %2").arg(t).arg(track->name()),1500);
}

/**
 * @short Show pattern properties of current pattern
 */
void TrackView::showPatternProperties(void){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->showProperties();
}

/**
 * @short Show controllers properties of current pattern
 */
void TrackView::showControllersProperties(void){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->showControllersProperties();
}

/**
 * @show Adds a note column in current Pattern
 */
void TrackView::addColumn(){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->addColumn();
  statusMessage("Added column",500);
}
/**
 * @short Deletes last note column in the current Pattern
 */
void TrackView::delColumn(){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->delColumn();
  statusMessage("Deleted column",500);
}
/**
 * @short Adds a controller to the current Pattern
 */
void TrackView::addController(){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->addController();
  statusMessage("Added controller",500);
}
/**
 * @short Deletes last controller from Pattern
 */
void TrackView::delController(){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->delController();
  statusMessage("Deleted controller",500);
}

/**
 * @short Changes the Rows and columns according to the spin boxes
 */
void TrackView::changeRowsAndColumns(){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv){
    pv->setRows(rowsSpin->value());
    pv->setCols(columnsSpin->value());
  }
  statusMessage("Changed topography",500);
}

void TrackView::addInitialSysEx(){
  int i,n=-1;
  KPopupMenu popup(this);
  popup.insertTitle(i18n("Select SysEx Message"));

  SysExBag *sysexbag=track->getSysExBag();
  for (i=0;i<signed(sysexbag->count());i++){
    n=popup.insertItem(sysexbag->at(i)->getName());
  }
  i=n+i-1;
  n=popup.exec(QCursor::pos());
  if (n==-1)
    return;

  n=i-n;
  // Finally n have the sysex number
  emit addInitialSysEx(sysexbag->at(n));
}

void TrackView::addedInitialSysEx(SysEx *s){
  initialSysEx->insertItem(s->getName());
  statusMessage(QString("Added initial SysEx %1").
                arg(s->getName()),500);
}


void TrackView::delSelectedInitialSysEx(){
  QString selected=initialSysEx->currentText();

  SysExBag *sysexbag=track->getSysExBag();
  unsigned int i;
  for (i=0;i<sysexbag->count();i++){
    if (selected==sysexbag->at(i)->getName()){
      emit delInitialSysEx(sysexbag->at(i));
      break;
    }
  }
}

void TrackView::deletedInitialSysEx(SysEx *s){
  initialSysEx->
    removeItem(initialSysEx->
         index(initialSysEx->
         findItem(s->getName())));
  statusMessage(QString("Deleted initial SysEx %1").
                arg(s->getName()),500);
}

void TrackView::delSelectedPattern(){
  int n=patternList->currentItem();

  if (-1==n)
    return;

  Pattern *pt=((PatternView*)patternView->widget(n))->getPattern();
  int ok=KMessageBox::questionYesNo(this,
                                    QString("Are you sure you want to delete"
                                            " Pattern <%1>?\n").
                                    arg(pt->name()),
                                    "Sure to delete?");
  if (ok!=KMessageBox::Yes)
    return;

  track->delPattern(pt);
  if (n>0)
    selectedPattern(n-1);
}

void TrackView::deletedPattern(Pattern *pt){
  PatternView *v;
  unsigned int i=0;
  v=(PatternView*)patternView->widget(i);
  while (v!=NULL){
    if (v->getPattern()==pt){

      // I remove the widget stack widget
      patternView->removeWidget(v);
      delete v;
      int n=i;
      i++;
      v=(PatternView*)patternView->widget(i);
      while (v!=NULL){
        patternView->removeWidget(v);
        patternView->addWidget(v,i-1);
        i++;
        v=(PatternView*)patternView->widget(i);
      }

      // Remove it from the list
      n=patternList->currentItem();
      patternList->clear();

      for (i=0;i<track->npatterns();i++){
        char letterId='*';
        if (i<10)
          letterId='0'+i;
        else
          letterId='a'+i-10;
        
        patternList->insertItem(QString("%1. %2").
              arg(letterId).
              arg(track->pattern(i)->name()),
              i);
      }
      patternList->setSelected(n,true);
      return;
    }
    i++;
    patternView->raiseWidget(i);
    v=(PatternView*)patternView->widget(i);
  }

  statusMessage("Deleted Pattern",1000);
}
    
/** 
  * @short Performs a copy of the selected region to the clipboard
  */
void TrackView::doCopy(){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->doCopy();
}

/**
 * @short Copies data from the clipboard to the current pattern
 */
void TrackView::doPaste(){
  PatternView *pv=(PatternView*)patternView->visibleWidget();
  if (NULL!=pv)
    pv->doPaste();
}
